import os
import sys
import re
import hashlib
import importlib.util
import platform
import toml
from pathlib import Path

from .snippets import NodeSnippetData
from . import logging, plugin_info, paths

from typing import List, Tuple, Dict, Any, Union, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .basenode import BaseNode


plugins = {}
presets: Dict[str, Dict[str, NodeSnippetData]] = {}
# map of node type -2->
#   preset_name -2->
#     dict of parameter name -2-> value
nodes_settings: Dict[str, Dict[str, Dict[str, Any]]] = {}
default_settings_config: Dict[str, str] = {}
__plugin_file_hashes = {}

# package is identified by it's path, but it's handy to address them by short names
# short name is generated from dir name. packages can have same dir names, but then
# only first one will get into this locations dict
__package_locations: Dict[str, Path] = {}

logger = logging.get_logger('plugin_loader')


def _install_node(filepath, plugin_category, parent_package=None):
    """

    :param filepath:
    :param plugin_category:
    :param parent_package: path to the base of the package, if this plugin is part of one, else - None
    :return:
    """
    filename = os.path.basename(filepath)
    filebasename, fileext = os.path.splitext(filename)

    modpath = f'lifeblood.nodeplugins.{plugin_category}.{filebasename}'
    mod_spec = importlib.util.spec_from_file_location(modpath, filepath)
    try:
        mod = importlib.util.module_from_spec(mod_spec)
        mod_spec.loader.exec_module(mod)
        mod._plugin_info = plugin_info.PluginInfo(filepath, parent_package)
    except:
        logger.exception(f'failed to load plugin "{filebasename}". skipping.')
        return
    for requred_attr in ('node_class',):
        if not hasattr(mod, requred_attr):
            logger.error(f'error loading plugin "{filebasename}". '
                         f'required method {requred_attr} is missing.')
            return
    plugins[mod.node_class().type_name()] = mod
    hasher = hashlib.md5()
    with open(filepath, 'rb') as f:
        hasher.update(f.read())
    __plugin_file_hashes[mod.node_class().type_name()] = hasher.hexdigest()
    sys.modules[modpath] = mod


def _install_package(package_path, plugin_category):
    """
    package structure:
        [package_name:dir]
        |_bin
        | |_any                 <- this is always added to PATH
        | |_system-arch1        <- these are added to PATH only if system+arch match
        | |_system-arch2        <-/
        |_python
        | |_X           <- these are added to PYTHONPATH based on X.Y
        | |_X.Y         <-/
        |_nodes
        | |_node1.py    <- these are loaded as usual node plugins
        | |_node2.py    <-/
        |_data          <- just a convenient place to store shit, can be accessed with data from plugin
        |_settings      <- for future saved nodes settings. not implemented yet
        | |_node_type_name1
        | | |_settings1.lbs
        | | |_settings2.lbs
        | |_node_type_name2
        | | |_settings1.lbs
        | | |_settings2.lbs
        |_whatever_file1.lol
        |_whatever_dir1
          |_whatever_file2.lol

    :param package_path:
    :param plugin_category:
    :return:
    """
    package_name = os.path.basename(package_path)
    global __package_locations
    if package_name not in __package_locations:  # read logic of this up
        __package_locations[package_name] = Path(package_path)
    # add extra bin paths
    extra_bins = []
    for subbin in (f'{platform.system().lower()}-{platform.machine().lower()}', 'any'):
        bin_base_path = os.path.join(package_path, 'bin', subbin)
        if not os.path.exists(bin_base_path):
            continue
        extra_bins.append(bin_base_path)
    if extra_bins:
        os.environ['PATH'] = os.pathsep.join(extra_bins) + os.environ['PATH']

    # install extra python modules
    python_base_path = os.path.join(package_path, 'python')
    if os.path.exists(python_base_path):
        sysver = sys.version_info
        pyvers = [tuple(int(y) for y in x.split('.')) for x in os.listdir(python_base_path) if x.isdigit() or re.match(r'^\d+\.\d+$', x)]
        pyvers = [x for x in pyvers if x[0] == sysver.major
                                       and (len(x) < 2 or x[1] == sysver.minor)
                                       and (len(x) < 3 or x[2] == sysver.micro)]
        pyvers = sorted(pyvers, key=lambda x: len(x), reverse=True)
        for pyver in pyvers:
            extra_python = os.path.join(python_base_path, '.'.join(str(x) for x in pyver))
            sys.path.append(extra_python)
            os.environ['PYTHONPATH'] = os.pathsep.join((extra_python, os.environ['PYTHONPATH'])) if 'PYTHONPATH' in os.environ else extra_python

    # install nodes
    nodes_path = os.path.join(package_path, 'nodes')
    if os.path.exists(nodes_path):
        for filename in os.listdir(nodes_path):
            filebasename, fileext = os.path.splitext(filename)
            if fileext != '.py':
                continue
            _install_node(os.path.join(nodes_path, filename), plugin_category, package_path)

    # install presets
    presets_path = os.path.join(package_path, 'presets')
    if os.path.exists(presets_path):
        for filename in os.listdir(presets_path):
            filebasename, fileext = os.path.splitext(filename)
            if fileext != '.lbp':
                continue
            try:
                with open(os.path.join(presets_path, filename), 'rb') as f:
                    snippet = NodeSnippetData.deserialize(f.read())
                snippet.add_tag('preset')
            except Exception as e:
                logger.error(f'failed to load snippet {filebasename}, error: {str(e)}')
                continue

            if package_name not in presets:
                presets[package_name] = {}
            presets[package_name][snippet.label] = snippet

    # install node settings
    settings_path = os.path.join(package_path, 'settings')
    if os.path.exists(settings_path):
        for nodetype_name in os.listdir(settings_path):
            if nodetype_name not in nodes_settings:
                nodes_settings[nodetype_name] = {}
            nodetype_path = os.path.join(settings_path, nodetype_name)
            for preset_filename in os.listdir(nodetype_path):
                preset_name, fileext = os.path.splitext(preset_filename)
                if fileext != '.lbs':
                    continue
                try:
                    with open(os.path.join(nodetype_path, preset_filename), 'r') as f:
                        nodes_settings[nodetype_name][preset_name] = toml.load(f)
                except Exception as e:
                    logger.error(f'failed to load settings {nodetype_name}/{preset_name}, error: {str(e)}')


def init():
    logger.info('loading core plugins')
    global plugins
    plugins = {}
    plugin_paths: List[Tuple[str, str]] = []  # list of tuples of path to dir, plugin category
    core_plugins_path = os.path.join(os.path.dirname(__file__), 'core_nodes')
    stock_plugins_path = os.path.join(os.path.dirname(__file__), 'stock_nodes')
    custom_plugins_path = paths.config_path('', 'custom_plugins')
    plugin_paths.append((core_plugins_path, 'core'))
    plugin_paths.append((stock_plugins_path, 'stock'))
    (custom_plugins_path/'custom_default').mkdir(parents=True, exist_ok=True)

    plugin_paths.append((str(custom_plugins_path), 'user'))

    extra_paths = []
    for path in os.environ.get('LIFEBLOOD_PLUGIN_PATH', '').split(os.pathsep):
        if path == '':
            continue
        if not os.path.isabs(path):
            logger.warning(f'"{path}" is not absolute, skipping')
            continue
        if not os.path.exists(path):
            logger.warning(f'"{path}" does not exist, skipping')
            continue
        extra_paths.append(path)
        logger.debug(f'using extra plugin path: "{path}"')

    plugin_paths.extend((x, 'extra') for x in extra_paths)

    for plugin_path, plugin_category in plugin_paths:
        for filename in os.listdir(plugin_path):
            filepath = os.path.join(plugin_path, filename)
            if os.path.isdir(filepath):
                _install_package(filepath, plugin_category)
            else:
                filebasename, fileext = os.path.splitext(filename)
                if fileext != '.py':
                    continue
                _install_node(filepath, plugin_category)

    logger.info('loaded node types:\n\t' + '\n\t'.join(plugins.keys()))
    logger.info('loaded node presets:\n\t' + '\n\t'.join(f'{pkg}::{label}' for pkg, pkgdata in presets.items() for label in pkgdata.keys()))

    # load default settings
    default_settings_config_path = paths.config_path('defaults.toml', 'scheduler.nodes')
    global default_settings_config
    if default_settings_config_path.exists():
        with open(default_settings_config_path) as f:
            default_settings_config = toml.load(f)

        bad_defaults = []
        for node_type, settings_name in default_settings_config.items():
            if settings_name not in nodes_settings.get(node_type, {}):
                logger.warning(f'"{settings_name}" is set as default for "{node_type}", but no such settings is loaded')
                bad_defaults.append(node_type)
                continue


def plugin_hash(plugin_name) -> str:
    return __plugin_file_hashes[plugin_name]


def add_settings_to_existing_package(package_name_or_path: Union[str, Path], node_type_name: str, settings_name: str, settings: Dict[str, Any]):

    if isinstance(package_name_or_path, str) and package_name_or_path in __package_locations:
        package_name_or_path = __package_locations[package_name_or_path]
    else:
        package_name_or_path = Path(package_name_or_path)
    if package_name_or_path not in __package_locations.values():
        raise RuntimeError('no package with that name or pathfound')

    # at this point package_name_or_path is path
    assert(package_name_or_path.exists())
    base_path = package_name_or_path / 'settings' / node_type_name
    if not base_path.exists():
        base_path.mkdir(parents=True, exist_ok=True)
    with open(base_path / (settings_name + '.lbs'), 'w') as f:
        toml.dump(settings, f)

    # add to settings
    nodes_settings.setdefault(node_type_name, {})[settings_name] = settings


def set_settings_as_default(node_type_name: str, settings_name: Optional[str]):
    """

    :param node_type_name:
    :param settings_name: if None - unset any defaults
    :return:
    """
    if node_type_name not in nodes_settings:
        raise RuntimeError(f'node type "{nodes_settings}" is unknown')
    if settings_name is not None and settings_name not in nodes_settings[node_type_name]:
        raise RuntimeError(f'node type "{nodes_settings}" doesn\'t have settings "{settings_name}"')
    if settings_name is None and node_type_name in default_settings_config:
        del default_settings_config[node_type_name]
    else:
        default_settings_config[node_type_name] = settings_name
    with open(paths.config_path('defaults.toml', 'scheduler.nodes'), 'w') as f:
        toml.dump(default_settings_config, f)


def create_node(type_name: str, name, scheduler_parent, node_id):  # type: (str, str, Scheduler, int) -> BaseNode
    """
    this function is a global node creation point.
    it has to be available somewhere global, so plugins loaded from dynamically created modules have an entry point for pickle
    """
    if type_name not in plugins:
        if type_name == 'basenode':  # debug case! base class should never be created directly!
            logger.warning('creating BASENODE. if it\'s not for debug/test purposes - it\'s bad!')
            from .basenode import BaseNode
            node = BaseNode(name)
        raise RuntimeError('unknown plugin')
    node = plugins[type_name].node_class()(name)
    # now set defaults, before parent is set to prevent ui callbacks to parent
    if type_name in default_settings_config:
        node.apply_settings(default_settings_config[type_name])
    node._set_parent(scheduler_parent, node_id)
    return node



from copy import copy
from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.processingcontext import ProcessingContext
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment
from lifeblood.text import filter_by_pattern

from lifeblood_stock_houdini_helpers.rop_base_node import RopBaseNode

from typing import Iterable, Optional


def node_class():
    return HipUsdGenerator


class HipUsdGenerator(RopBaseNode):
    @classmethod
    def label(cls) -> str:
        return 'usd generator'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'hip', 'houdini', 'usd', 'lop', 'solaris', 'generator', 'render', 'stock'

    @classmethod
    def type_name(cls) -> str:
        return 'hip_usd_generator'

    def __init__(self, name):
        super(HipUsdGenerator, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.5, 0.25, 0.125)

            ui.parameter('scene file output').set_value("`config['global_scratch_location']`/`node.name`/`task.name`_`task.id`/usd/`node.name`_$F4")

            skip_param = ui.parameter('skip if exists')
            skip_param.set_value(False)
            skip_param.set_locked(True)
            skip_param.set_hidden(True)

            ui.add_separator()

            with ui.parameters_on_same_line_block():
                use_custom = ui.add_parameter('use custom usd attr for img', 'Custom USD prim/attr for output image', NodeParameterType.BOOL, False)
                ui.add_parameter('custom usd attr for img', None, NodeParameterType.STRING, '/Render/Products/renderproduct/productName')

    # that below - for mantra
    #
    # def _take_parm_name(self, context) -> str:
    #     return 'take'
    #
    # def _parms_to_set_before_render(self, context) -> dict:
    #     return {'soho_outputmode': 1,
    #             'vm_inlinestorage': int(context.param_value('ifd force inline'))}
    #
    # def _scene_file_parm_name(self, context) -> str:
    #     return 'soho_diskfile'
    #
    # def _image_path_parm_name(self, context) -> str:
    #     return 'vm_picture'

    def _take_parm_name(self, context) -> str:
        return 'take'

    def _parms_to_set_before_render(self, context) -> dict:
        return {'mkpath': True,
                'runcommand': False,
                'husk_mplay': False
                }

    def _scene_file_getting_code(self, context) -> Optional[str]:
        return '\n'.join([f'base = node.evalParm("savetodirectory_directory")',
                          f'filename_parm = node.parm("lopoutput")',
                          f'if filename_parm is not None:',
                          f'    filename = filename_parm.eval()',
                          f'else:',
                          f'    filename = "render.usd"',
                          f'if os.path.splitext(filename) == ".usd" and hou.licenseCategory() in (hou.licenseCategoryType.Apprentice, hou.licenseCategoryType.ApprenticeHD):',
                          f'    filename += "nc"',
                          f'return os.path.join(base, filename)'])

    def _scene_file_parm_name(self, context) -> str:
        return 'savetodirectory_directory'

    def _image_getting_code(self, context: ProcessingContext) -> Optional[str]:
        usd_attr_path = context.param_value('custom usd attr for img') if context.param_value('use custom usd attr for img') else '/Render/Products/renderproduct/productName'
        prim_path, attr_name = usd_attr_path.rsplit('/', 1)
        return '\n'.join([f'override = node.parm("outputimage").evalAsString().strip()',
                          f'if override:',
                          f'    return override',
                          f'prim = node.input(0).stage().GetPrimAtPath({repr(prim_path)})',
                          f'attr = prim.GetAttribute({repr(attr_name)})',
                          f'return attr.Get(frame)'])

    def _node_specialization_code(self, context):
        return '\n'.join([f'if node.type().name() == "karma":',
                          f'    if node.type().category() == hou.nodeTypeCategories()["Driver"]:',
                          f'        return node.node("lopnet/karma/rop_usdrender")',
                          f'    elif node.type().category() == hou.nodeTypeCategories()["Lop"]:',
                          f'        return node.node("rop_usdrender")',
                          f'return node'])

    def _image_path_parm_name(self, context) -> str:
        return ''  # value ignored cuz _image_getting_code is returning non None



import asyncio
import uuid
import struct
from io import BytesIO
from .messages import MessageInterface, Message
from .message_stream import MessageSendStreamBase, MessageReceiveStreamBase
from .enums import MessageType
from .address import AddressChain, DirectAddress
from .exceptions import MessageReceivingError, MessageSendingError, MessageTransferTimeoutError, NoMessageError
from .defaults import default_stream_timeout

from typing import Optional, Tuple, Union


def _write_header(stream: asyncio.StreamWriter, data_size: int, message_type: MessageType, source: AddressChain, destination: AddressChain, session: Optional[uuid.UUID]):
    src_data = source.encode('UTF-8')
    src_data_size = len(src_data)
    dest_data = destination.encode('UTF-8')
    dest_data_size = len(dest_data)
    has_session = session is not None
    stream.write(struct.pack('>QQQ?', data_size + 16 + src_data_size + dest_data_size + 1 + message_type.message_type_size() + (16 if has_session else 0),
                             src_data_size,
                             dest_data_size,
                             session is not None))
    stream.write(message_type.value)
    stream.write(src_data)
    stream.write(dest_data)
    if has_session:
        stream.write(session.bytes)


async def _read_header(stream: asyncio.StreamReader, timeout: float) -> Tuple[int, MessageType, AddressChain, AddressChain, Optional[uuid.UUID]]:
    return await asyncio.wait_for(__read_header(stream), timeout=timeout)


async def __read_header(stream: asyncio.StreamReader) -> Tuple[int, MessageType, AddressChain, AddressChain, Optional[uuid.UUID]]:
    """
    returns: (data_size, destination, session)
    """
    message_size, src_data_size, dest_data_size, has_session = struct.unpack('>QQQ?', await stream.readexactly(25))
    message_type = MessageType(await stream.readexactly(MessageType.message_type_size()))
    source = (await stream.readexactly(src_data_size)).decode('UTF-8')
    destination = (await stream.readexactly(dest_data_size)).decode('UTF-8')
    session = uuid.UUID(bytes=await stream.readexactly(16)) if has_session else None
    return message_size - 8-src_data_size - 8-dest_data_size - 1 - MessageType.message_type_size() - (16 if has_session else 0), \
           message_type, \
           AddressChain(source), \
           AddressChain(destination), \
           session


async def _from_stream_reader(stream: asyncio.StreamReader):
    data_size, message_type, source, destination, session = await _read_header(stream)
    data = await stream.readexactly(data_size)
    return Message(data, message_type, source, destination, session)


async def _serialize_to_stream_writer(message: Message, stream: asyncio.StreamWriter):
    _write_header(stream,
                  len(message.message_body()),
                  message.message_type(),
                  message.message_source(),
                  message.message_destination(),
                  message.message_session())
    stream.write(message.message_body())
    await stream.drain()


class WriterStreamRawMessageWrapper(MessageInterface):
    """
    wraps writes into "messages" that have predictable borders, that can be redirected easily
    """
    def __init__(self, writer: asyncio.StreamWriter, *, message_type: Optional[MessageType] = None, source: AddressChain, destination: AddressChain, session: Optional[uuid.UUID] = None):
        """
        Note: source and destination must be provided:
            while destination is basically where reader/writer are connected,
            source is where reply should come, so it's NOT current socket's address

        """
        self.__writer = writer
        self.__buffer: Optional[BytesIO] = None
        self.__size = 0
        self.__source: AddressChain = source
        self.__destination: AddressChain = destination
        self.__session = session
        self.__message_type: MessageType = message_type or MessageType.DEFAULT_MESSAGE
        self.__initialized = False

    def initialized(self):
        return self.__initialized

    async def __aenter__(self) -> "WriterStreamRawMessageWrapper":
        if self.__initialized:
            raise RuntimeError('message was already initialized')
        self.__initialized = True
        self.__buffer = BytesIO()
        self.__size = 0
        return self

    def write(self, data: bytes):
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        self.__size += len(data)
        self.__buffer.write(data)

    def write_string(self, s: str):
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        b = s.encode('UTF-8')
        self.write(struct.pack('>Q', len(b)))
        self.write(b)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        _write_header(self.__writer, self.__size, self.__message_type, self.__source, self.__destination, self.__session)
        self.__writer.write(self.__buffer.getbuffer())
        await self.__writer.drain()

    def message_size(self) -> int:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__size

    def message_body(self) -> Union[bytes, memoryview]:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__buffer.getbuffer()

    def message_destination(self) -> AddressChain:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__destination

    def message_source(self) -> AddressChain:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__source

    def message_session(self) -> Optional[uuid.UUID]:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__session

    def message_type(self) -> MessageType:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__message_type

    def to_message(self) -> Message:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return Message(self.message_body(), self.message_type(), self.message_source(), self.message_destination(), self.message_session())


class ReaderStreamRawMessageWrapper(MessageInterface):
    def __init__(self, reader: asyncio.StreamReader, session: Optional[uuid.UUID] = None, timeout: Optional[float] = None):
        self.__reader = reader
        self.__message_body_size: int = 0
        self.__already_read = 0
        self.__source: Optional[AddressChain] = None
        self.__destination: Optional[AddressChain] = None
        self.__message_type: Optional[MessageType] = None
        self.__initialized = False
        self.__buffer = BytesIO()
        self.__session = session
        self.__timeout = timeout

    def total_read_bytes(self) -> int:
        return self.__already_read

    def initialized(self):
        return self.__initialized

    def message_size(self):
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__message_body_size

    def message_body(self) -> Union[bytes, memoryview]:
        """
        message body so far. this ONLY guaranteed returns the whole message body after with block end
        """
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__buffer.getbuffer()

    def message_destination(self) -> AddressChain:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__destination

    def message_source(self) -> AddressChain:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__source

    def message_session(self) -> Optional[uuid.UUID]:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__session

    def message_type(self) -> MessageType:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return self.__message_type

    async def __aenter__(self) -> "ReaderStreamRawMessageWrapper":
        if self.__initialized:
            raise RuntimeError('message was already initialized')
        self.__initialized = True
        self.__message_body_size, self.__message_type, self.__source, self.__destination, session = await _read_header(self.__reader, timeout=self.__timeout)
        if self.__session is not None:  # if we expect specific session
            if session != self.__session:
                raise RuntimeError('received message has wrong session')
        else:
            self.__session = session
        self.__already_read = 0
        return self

    async def readexactly(self, size: int):
        if size < 0:
            return
        elif size == 0:
            return b''
        data = await asyncio.wait_for(self.__reader.readexactly(size), timeout=self.__timeout)
        self.__already_read += size
        self.__buffer.write(data)
        return data

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.readexactly(self.__message_body_size - self.__already_read)

    def to_message(self) -> Message:
        if not self.__initialized:
            raise RuntimeError('message was not initialized')
        return Message(self.message_body(), self.message_type(), self.message_source(), self.message_destination(), self.message_session())


class MessageSendStream(MessageSendStreamBase):
    def __init__(self,
                 reader_stream: asyncio.StreamReader,
                 writer_stream: asyncio.StreamWriter,
                 *,
                 reply_address: DirectAddress,
                 destination_address: DirectAddress,
                 stream_timeout: float = default_stream_timeout,
                 confirmation_timeout: float = default_stream_timeout):
        super().__init__(reply_address, destination_address, stream_timeout=stream_timeout, message_confirmation_timeout=confirmation_timeout)
        self.__reader = reader_stream
        self.__writer = writer_stream

    # def start_new_message(self, destination: AddressChain, session: uuid.UUID) -> WriterStreamRawMessageWrapper:
    #     return WriterStreamRawMessageWrapper(self.__writer, source=self.reply_address(), destination=destination, session=session)

    async def _send_message_implementation(self, message: Message, timeout: Optional[float]):
        """
        override this with actual message sending implementation
        """
        await _serialize_to_stream_writer(message, self.__writer)
        await self.__writer.drain()  # TODO: timeout not implemented

    async def _ack_message_implementation(self, message: Message, timeout: Optional[float]):
        try:
            if not struct.unpack('?', await asyncio.wait_for(self.__reader.readexactly(1),
                                                             timeout=timeout))[0]:
                raise MessageSendingError('message delivery failed', wrapped_exception=None)
        except asyncio.TimeoutError as e:
            raise MessageTransferTimeoutError(wrapped_exception=e) from None

    def close(self):
        if not self.__writer.is_closing():
            self.__writer.close()

    async def wait_closed(self):
        await self.__writer.wait_closed()


class MessageReceiveStream(MessageReceiveStreamBase):
    def __init__(self, reader_stream: asyncio.StreamReader, writer_stream: asyncio.StreamWriter, *, this_address: DirectAddress, other_end_address: DirectAddress, stream_timeout: float = default_stream_timeout):
        super().__init__(this_address, other_end_address, stream_timeout=stream_timeout)
        self.__reader = reader_stream
        self.__writer = writer_stream

    # def start_new_message(self, destination: AddressChain, session: uuid.UUID) -> WriterStreamRawMessageWrapper:
    #     return WriterStreamRawMessageWrapper(self.__writer, source=self.reply_address(), destination=destination, session=session)

    def __start_receiving_message(self):
        return ReaderStreamRawMessageWrapper(self.__reader, timeout=self._stream_timeout())

    async def _receive_message_implementation(self) -> Message:
        """
        override this with actual message receiving implementation
        """
        stream = None
        try:
            async with self.__start_receiving_message() as stream:
                pass  # will receive all
            return stream.to_message()
        except asyncio.TimeoutError as e:
            if stream is not None:
                if stream.total_read_bytes() == 0:
                    raise NoMessageError(wrapped_exception=e) from None
            raise MessageTransferTimeoutError(f'read {stream.total_read_bytes()}: {stream.message_body().decode("latin1")}', wrapped_exception=e) from None

    async def _acknowledge_message_implementation(self, status: bool):
        self.__writer.write(struct.pack('?', status))
        await self.__writer.drain()

    def close(self):
        self.__writer.close()

    async def wait_closed(self):
        await self.__writer.wait_closed()



import io
import socket
import struct
from .exceptions import IncompleteReadError

from typing import Tuple


class BufferedConnection:
    def __init__(self, address, buffering=None, timeout=30):
        """

        :param address:
        :param buffering: None means default buffer size, 0 means unbuffered
        :return:
        """
        self.__sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__sock.settimeout(timeout)
        self.__sock.connect(address)

        if buffering is None:
            buffering = -1
        if buffering < 0:
            buffering = io.DEFAULT_BUFFER_SIZE

        reader = self.__sock.makefile('rb', buffering)
        writer = self.__sock.makefile('wb', buffering)
        assert isinstance(reader, io.BufferedReader)
        assert isinstance(writer, io.BufferedWriter)
        self.__reader = BufferedReaderWrapper(reader, buffering)
        self.__writer = BufferedWriterWrapper(writer, buffering)

    def close(self):
        self.__writer.flush()
        self.__writer.close()
        self.__reader.close()
        self.__sock.close()

    @property
    def reader(self) -> "BufferedReaderWrapper":
        return self.__reader

    @property
    def writer(self) -> "BufferedWriterWrapper":
        return self.__writer

    def get_rw_pair(self) -> Tuple["BufferedReaderWrapper", "BufferedWriterWrapper"]:
        return self.__reader, self.__writer


class BufferedReader(io.BufferedReader):
    def readexactly(self, size) -> bytes:
        # a tiny optimization: first try to read with a single call, and then employ general tactics if that fails
        part = self.read(size)
        lpart = len(part)
        if lpart == 0:
            raise IncompleteReadError(f'wanted to read {size}, but got EOF')
        if lpart == size:
            return part

        # and now generic algorithm
        parts = [part]
        total_read = lpart
        while total_read < size:
            part = self.read(size)
            lpart = len(part)
            if lpart == 0:  # meaning EOF
                raise IncompleteReadError(f'wanted to read {size}, but got only {total_read} before EOF')
            total_read += lpart
            parts.append(part)
        return b''.join(parts)


class BufferedWriter(io.BufferedWriter):
    pass


class BufferedReaderWrapper(BufferedReader):
    def __init__(self, stuff_to_wrap: io.BufferedReader, buffering=None):
        if buffering is None:
            buffering = -1
        if buffering < 0:
            buffering = io.DEFAULT_BUFFER_SIZE
        super().__init__(stuff_to_wrap.raw, buffering)
        self.__wrapped = stuff_to_wrap

    def read_string(self):
        data_len, = struct.unpack('>Q', self.readexactly(8))
        if data_len == 0:
            return ''
        return self.readexactly(data_len).decode('UTF-8')

    def close(self) -> None:
        super().close()
        self.__wrapped.close()


class BufferedWriterWrapper(BufferedWriter):
    def __init__(self, stuff_to_wrap: io.BufferedWriter, buffering=None):
        super().__init__(stuff_to_wrap.raw, buffering)
        self.__wrapped = stuff_to_wrap

    def write_string(self, text: str):
        data = text.encode('UTF-8')
        self.write(struct.pack('>Q', len(data)))
        self.write(data)

    def close(self) -> None:
        super().close()
        self.__wrapped.close()



import asyncio
from enum import Enum
import pickle
import json
import re
from copy import copy, deepcopy
from typing import Dict, Optional, List, Any
from .nodethings import ProcessingResult, ProcessingError
from .uidata import NodeUi, ParameterNotFound, ParameterReadonly, ParameterLocked, ParameterCannotHaveExpressions, Parameter
from .pluginloader import create_node, plugin_hash, nodes_settings
from .processingcontext import ProcessingContext
from .logging import get_logger
from .enums import NodeParameterType, WorkerType
from .plugin_info import PluginInfo

from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from .scheduler import Scheduler
    from logging import Logger


class BaseNode:
    @classmethod
    def label(cls) -> str:
        raise NotImplementedError()

    @classmethod
    def tags(cls) -> Iterable[str]:
        raise NotImplementedError()

    @classmethod
    def type_name(cls) -> str:
        raise NotImplementedError()

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def __init__(self, name: str):
        self.__parent: Scheduler = None
        self.__parent_nid: int = None
        self._parameters: NodeUi = NodeUi(self)
        self.__name = name
        try:
            mytype = self.type_name()
        except NotImplementedError:
            mytype = None
        self.__logger = get_logger(f'BaseNode.{mytype}' if mytype is not None else 'BaseNode')
        # subclass is expected to add parameters at this point

    def _set_parent(self, parent_scheduler, node_id):
        self.__parent = parent_scheduler
        self.__parent_nid = node_id

    def logger(self) -> "Logger":
        return self.__logger

    def name(self):
        return self.__name

    def set_name(self, name: str):
        self.__name = name

    def id(self):
        return self.__parent_nid

    # # MAYBE UNCOMMENT THESE WHEN ALL NODES ARE REFACTORED TO USE CONTEXT?
    # def param_value(self, param_name, context: Optional[ProcessingContext] = None) -> Any:
    #     """
    #     shortcut to node.get_ui().parameter_value
    #     :param param_name:
    #     :param context: context in which to evaluate expression
    #     :return:
    #     """
    #     return self._parameters.parameter(param_name).value(context)
    #
    def param(self, param_name) -> Parameter:
        """
        shortcut to node.get_ui().parameter

        :param param_name:
        :return:
        """
        return self._parameters.parameter(param_name)

    def set_param_value(self, param_name, param_value) -> None:
        """
        shortcut to node.get_ui().set_parameter
        :param param_name: parameter name
        :param param_value: value to set
        :return:
        """
        return self._parameters.parameter(param_name).set_value(param_value)

    def get_ui(self) -> NodeUi:
        return self._parameters

    def is_input_connected(self, input_name: str) -> bool:
        """
        returns wether or not specified input is connected to the node
        note that these methods are supposed to be called both from main thread AND from within executor pool thread
        so creating tasks becomes tricky.
        :param input_name:
        :return:
        """
        try:
            thread_loop = asyncio.get_running_loop()
        except RuntimeError:  # no loop
            thread_loop = None
        fut = asyncio.run_coroutine_threadsafe(self.__parent.get_node_input_connections(self.__parent_nid, input_name), self.__parent.get_event_loop())
        if thread_loop is self.__parent.get_event_loop():  # we are in scheduler's main loop
            self.__logger.error('this method cannot be called from the main scheduler thread')
            raise RuntimeError('this method cannot be called from the main scheduler thread')
        else:
            conns = fut.result(60)
        return len(conns) > 0

    def is_output_connected(self, output_name: str):
        """
        returns wether or not specified output is connected to the node
        :param output_name:
        :return:
        """
        try:
            thread_loop = asyncio.get_running_loop()
        except RuntimeError:  # no loop
            thread_loop = None
        fut = asyncio.run_coroutine_threadsafe(self.__parent.get_node_output_connections(self.__parent_nid, output_name), self.__parent.get_event_loop())
        if thread_loop is self.__parent.get_event_loop():  # we are in scheduler's main loop
            self.__logger.error('this method cannot be called from the main scheduler thread')
            raise RuntimeError('this method cannot be called from the main scheduler thread')
        else:
            conns = fut.result(60)
        return len(conns) > 0

    def _ui_changed(self, definition_changed=False):
        """
        this methods gets called by self and NodeUi when a parameter changes to trigger node's database update
        :return:
        """
        if self.__parent is not None:
            # it is important to create task in main scheduler's event loop
            try:
                thread_loop = asyncio.get_running_loop()
            except RuntimeError:  # no loop
                thread_loop = None
            fut = asyncio.run_coroutine_threadsafe(self.__parent.node_reports_changes_needs_saving(self.__parent_nid), self.__parent.get_event_loop())
            if thread_loop is self.__parent.get_event_loop():  # we are in scheduler's main loop
                self.__logger.warning('this method probably should not be called from the main scheduler thread')
            else:  # we are not in main thread
                fut.result(60)  # ensure callback is completed before continuing. not sure how much it's needed, but it feels safer
                # TODO: even though timeout even here is impossible in any sane situation - still we should do something about it
            # and this is a nono: asyncio.get_event_loop().create_task(self.__parent.node_reports_changes_needs_saving(self.__parent_nid))

    # def _state_changed(self):
    #     """
    #     this methods should be called when important node state was changes to trigger node's database update
    #     :return:
    #     """
    #     if self.__parent is not None:
    #         # TODO: note that this may happen at any point in processing,
    #         #  so IF node subclass has internal state - it has to deal with ensuring state is consistent at time or writing
    #         #  this may also apply to _ui_changed above, but nodes really SHOULD NOT change their own parameters during processing
    #         asyncio.get_event_loop().create_task(self.__parent.node_reports_changes_needs_saving(self.__parent_nid))

    def ready_to_process_task(self, task_dict) -> bool:
        """
        must be VERY fast - this will be run in main thread
        there is no context wrapper just to skip unnessesary overhead as not many nodes will ever override this
        """
        return True

    def ready_to_postprocess_task(self, task_dict) -> bool:
        """
        must be VERY fast - this will be run in main thread
        there is no context wrapper just to skip unnessesary overhead as not many nodes will ever override this
        """
        return True

    def _process_task_wrapper(self, task_dict) -> ProcessingResult:
        # with self.get_ui().lock_interface_readonly():  # TODO: this is bad, RETHINK!
        #  TODO: , in case threads do l1---r1    - release2 WILL leave lock in locked state forever, as it remembered it at l2
        #  TODO:                         l2---r2
        return self.process_task(ProcessingContext(self, task_dict))

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        """
        custom node classes subclassing this are supposed to override this method

        :param context:
        :return:
        """
        raise NotImplementedError()

    def _postprocess_task_wrapper(self, task_dict) -> ProcessingResult:
        # with self.get_ui().lock_interface_readonly():  #TODO: read comment for _process_task_wrapper
        return self.postprocess_task(ProcessingContext(self, task_dict))

    def postprocess_task(self, context: ProcessingContext) -> ProcessingResult:
        """
        custom node classes subclassing this are supposed to override this method

        :param context:
        :return:
        """
        return ProcessingResult()

    def copy_ui_to(self, to_node: "BaseNode"):
        newui = deepcopy(self._parameters)  # nodeUI redefines deepcopy to detach new copy from node
        to_node._parameters = newui
        newui.attach_to_node(to_node)

    def apply_settings(self, settings_name: str) -> None:
        mytype = self.type_name()
        if mytype not in nodes_settings:
            raise RuntimeError(f'no settings found for "{mytype}"')
        if settings_name not in nodes_settings[mytype]:
            raise RuntimeError(f'requested settings "{settings_name}" not found for type "{mytype}"')
        settings = nodes_settings[mytype][settings_name]
        with self.get_ui().postpone_ui_callbacks():
            for param_name, value in settings.items():
                try:
                    param = self.param(param_name)
                    if isinstance(value, dict):
                        if 'value' in value:
                            param.set_value(value['value'])
                        if 'expression' in value:
                            param.set_expression(value['expression'])
                    else:
                        if param.has_expression():
                            param.remove_expression()
                        param.set_value(value)
                except ParameterNotFound:
                    self.logger().warning(f'applying settings "{settings_name}": skipping unrecognized parameter "{param_name}"')
                    continue
                except ValueError as e:
                    self.logger().warning(f'applying settings "{settings_name}": skipping parameter "{param_name}": bad value type: {str(e)}')
                    continue

    # # some helpers
    # def _get_task_attributes(self, task_row):
    #     return json.loads(task_row.get('attributes', '{}'))

    #
    # Plugin info
    #
    @classmethod
    def my_plugin(cls) -> Optional[PluginInfo]:
        from . import pluginloader
        type_name = cls.type_name()
        if type_name not in pluginloader.plugins:
            return None
        return pluginloader.plugins[type_name]._plugin_info

    #
    # Serialize and back
    #
    def __reduce__(self):
        # typename = type(self).__module__
        # if '.' in typename:
        #     typename = typename.rsplit('.', 1)[-1]
        return create_node, (self.type_name(), '', None, None), self.__getstate__()

    def __getstate__(self):
        # TODO: if u ever implement parameter expressions - be VERY careful with pickling expressions referencing across nodes
        d = copy(self.__dict__)
        assert '_BaseNode__parent' in d
        d['_BaseNode__parent'] = None
        d['_BaseNode__parent_nid'] = None
        d['_BaseNode__saved_plugin_hash'] = plugin_hash(self.type_name())  # we will use this hash to detect plugin module changes on load
        return d

    def __setstate__(self, state):
        # the idea here is to update node's class instance IF plugin hash is different from the saved one
        # the hash being different means that node's definition was updated - we don't know how
        # so what we do is save all parameter values, merge old state values with new
        # and hope for the best...

        hash = plugin_hash(self.type_name())
        if hash != state.get('_BaseNode__saved_plugin_hash', None):
            self.__init__(state.get('name', ''))
            # update all except ui
            try:
                if '_parameters' in state:
                    old_ui: NodeUi = state['_parameters']
                    del state['_parameters']
                    self.__dict__.update(state)
                    new_ui = self.get_ui()
                    for param in old_ui.parameters():
                        try:
                            newparam = new_ui.parameter(param.name())
                        except ParameterNotFound:
                            continue
                        try:
                            newparam.set_value(param.unexpanded_value())
                        except ParameterReadonly:
                            newparam._Parameter__value = param.unexpanded_value()
                        except ParameterLocked:
                            newparam.set_locked(False)
                            newparam.set_value(param.unexpanded_value())
                            newparam.set_locked(True)
                        if param.has_expression():
                            try:
                                newparam.set_expression(param.expression())
                            except ParameterCannotHaveExpressions:
                                pass
                else:
                    self.__dict__.update(state)
            except AttributeError:
                # something changed so much that some core attrs are different
                get_logger('BaseNode').exception(f'could not update interface for some node of type {self.type_name()}. resetting node\'s inrerface')


            # TODO: if and whenever expressions are introduced - u need to take care of expressions here too!
        else:
            self.__dict__.update(state)

    def serialize(self) -> bytes:
        """
        by default we just serialize
        :return:
        """
        return pickle.dumps(self)

    async def serialize_async(self) -> bytes:
        return await asyncio.get_event_loop().run_in_executor(None, self.serialize)

    @classmethod
    def deserialize(cls, data: bytes, parent_scheduler, node_id):
        newobj = pickle.loads(data)
        newobj.__parent = parent_scheduler
        newobj.__parent_nid = node_id
        return newobj

    @classmethod
    async def deserialize_async(cls, data: bytes, parent_scheduler, node_id):
        return await asyncio.get_event_loop().run_in_executor(None, cls.deserialize, data, parent_scheduler, node_id)


class BaseNodeWithTaskRequirements(BaseNode):
    def __init__(self, name: str):
        super(BaseNodeWithTaskRequirements, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            with ui.collapsable_group_block('main worker requirements', 'worker requirements'):
                ui.add_parameter('priority adjustment', 'priority adjustment', NodeParameterType.FLOAT, 0).set_slider_visualization(-100, 100)
                with ui.parameters_on_same_line_block():
                    ui.add_parameter('worker cpu cost', 'min <cpu (cores)> preferred', NodeParameterType.FLOAT, 1.0).set_value_limits(value_min=0)
                    ui.add_parameter('worker cpu cost preferred', None, NodeParameterType.FLOAT, 0.0).set_value_limits(value_min=0)
                with ui.parameters_on_same_line_block():
                    ui.add_parameter('worker mem cost', 'min <memory (GBs)> preferred', NodeParameterType.FLOAT, 0.5).set_value_limits(value_min=0)
                    ui.add_parameter('worker mem cost preferred', None, NodeParameterType.FLOAT, 0.0).set_value_limits(value_min=0)
                ui.add_parameter('worker groups', 'groups (space or comma separated)', NodeParameterType.STRING, '')
                ui.add_parameter('worker type', 'worker type', NodeParameterType.INT, WorkerType.STANDARD.value)\
                    .add_menu((('standard', WorkerType.STANDARD.value),
                               ('scheduler helper', WorkerType.SCHEDULER_HELPER.value)))
                with ui.collapsable_group_block('gpu main worker requirements', 'gpu requirements'):
                    with ui.parameters_on_same_line_block():
                        ui.add_parameter('worker gpu cost', 'min <gpus> preferred', NodeParameterType.FLOAT, 0.0).set_value_limits(value_min=0)
                        ui.add_parameter('worker gpu cost preferred', None, NodeParameterType.FLOAT, 0.0).set_value_limits(value_min=0)
                    with ui.parameters_on_same_line_block():
                        ui.add_parameter('worker gpu mem cost', 'min <memory (GBs)> preferred', NodeParameterType.FLOAT, 0.0).set_value_limits(value_min=0)
                        ui.add_parameter('worker gpu mem cost preferred', None, NodeParameterType.FLOAT, 0.0).set_value_limits(value_min=0)

    def __apply_requirements(self, task_dict: dict, result: ProcessingResult):
        if result.invocation_job is not None:
            context = ProcessingContext(self, task_dict)
            raw_groups = context.param_value('worker groups').strip()
            reqs = result.invocation_job.requirements()
            if raw_groups != '':
                reqs.add_groups(re.split(r'[ ,]+', raw_groups))
            reqs.set_min_cpu_count(context.param_value('worker cpu cost'))
            reqs.set_min_memory_bytes(context.param_value('worker mem cost') * 10**9)
            reqs.set_min_gpu_count(context.param_value('worker gpu cost'))
            reqs.set_min_gpu_memory_bytes(context.param_value('worker gpu mem cost') * 10**9)
            # preferred
            pref_cpu_count = context.param_value('worker cpu cost preferred')
            pref_mem_bytes = context.param_value('worker mem cost preferred') * 10**9
            pref_gpu_count = context.param_value('worker gpu cost preferred')
            pref_gpu_mem_bytes = context.param_value('worker gpu mem cost preferred') * 10**9
            if pref_cpu_count > 0:
                reqs.set_preferred_cpu_count(pref_cpu_count)
            if pref_mem_bytes > 0:
                reqs.set_preferred_memory_bytes(pref_mem_bytes)
            if pref_gpu_count > 0:
                reqs.set_preferred_gpu_count(pref_gpu_count)
            if pref_gpu_mem_bytes > 0:
                reqs.set_preferred_gpu_memory_bytes(pref_gpu_mem_bytes)

            reqs.set_worker_type(WorkerType(context.param_value('worker type')))
            result.invocation_job.set_requirements(reqs)
            result.invocation_job.set_priority(context.param_value('priority adjustment'))
        return result

    def _process_task_wrapper(self, task_dict) -> ProcessingResult:
        result = super(BaseNodeWithTaskRequirements, self)._process_task_wrapper(task_dict)
        return self.__apply_requirements(task_dict, result)

    def _postprocess_task_wrapper(self, task_dict) -> ProcessingResult:
        result = super(BaseNodeWithTaskRequirements, self)._postprocess_task_wrapper(task_dict)
        return self.__apply_requirements(task_dict, result)


# class BaseNodeWithEnvironmentRequirements(BaseNode):
#     def __init__(self, name: str):
#         super(BaseNodeWithEnvironmentRequirements, self).__init__(name)
#         ui = self.get_ui()
#         with ui.initializing_interface_lock():
#             with ui.collapsable_group_block('main environment resolver', 'task environment resolver additional requirements'):
#                 ui.add_parameter('main env resolver name', 'resolver name', NodeParameterType.STRING, 'StandardEnvironmentResolver')
#                 with ui.multigroup_parameter_block('main env resolver arguments'):
#                     with ui.parameters_on_same_line_block():
#                         type_param = ui.add_parameter('main env resolver arg type', '', NodeParameterType.INT, 0)
#                         type_param.add_menu((('int', NodeParameterType.INT.value),
#                                              ('bool', NodeParameterType.BOOL.value),
#                                              ('float', NodeParameterType.FLOAT.value),
#                                              ('string', NodeParameterType.STRING.value),
#                                              ('json', -1)
#                                              ))
#
#                         ui.add_parameter('main env resolver arg svalue', 'val', NodeParameterType.STRING, '').append_visibility_condition(type_param, '==', NodeParameterType.STRING.value)
#                         ui.add_parameter('main env resolver arg ivalue', 'val', NodeParameterType.INT, 0).append_visibility_condition(type_param, '==', NodeParameterType.INT.value)
#                         ui.add_parameter('main env resolver arg fvalue', 'val', NodeParameterType.FLOAT, 0.0).append_visibility_condition(type_param, '==', NodeParameterType.FLOAT.value)
#                         ui.add_parameter('main env resolver arg bvalue', 'val', NodeParameterType.BOOL, False).append_visibility_condition(type_param, '==', NodeParameterType.BOOL.value)
#                         ui.add_parameter('main env resolver arg jvalue', 'val', NodeParameterType.STRING, '').append_visibility_condition(type_param, '==', -1)
#
#     def _process_task_wrapper(self, task_dict) -> ProcessingResult:
#         result = super(BaseNodeWithEnvironmentRequirements, self)._process_task_wrapper(task_dict)
#         result.invocation_job.environment_resolver_arguments()
#         return result



from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment
from lifeblood.text import filter_by_pattern

from typing import Iterable


def node_class():
    return HipScript


class HipScript(BaseNodeWithTaskRequirements):
    @classmethod
    def label(cls) -> str:
        return 'hip script executor'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'hip', 'houdini', 'script', 'python', 'hou'

    @classmethod
    def type_name(cls) -> str:
        return 'hip_script'

    @classmethod
    def description(cls) -> str:
        return 'opens hip, executes script and saves hip to the same or new location\n'

    def __init__(self, name):
        super(HipScript, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.5, 0.25, 0.125)
            ui.add_parameter('hip path', 'hip path', NodeParameterType.STRING, "`task['hipfile']`")
            with ui.parameters_on_same_line_block():
                mask_hip = ui.add_parameter('mask as different hip', 'mask as different hip file', NodeParameterType.BOOL, False)
                ui.add_parameter('mask hip path', '', NodeParameterType.STRING, "`task.get('hipfile_orig', task['hipfile'])`")\
                    .append_visibility_condition(mask_hip, '==', True)
            ui.add_parameter('script', 'script', NodeParameterType.STRING, '').set_text_multiline(syntax_hint='python')

            with ui.parameters_on_same_line_block():
                save_hip = ui.add_parameter('save different hip', 'save resulted file to a different location', NodeParameterType.BOOL, False)
                ui.add_parameter('save hip path', '', NodeParameterType.STRING, '')\
                    .append_visibility_condition(save_hip, '==', True)

            ui.parameter('worker type').set_hidden(True)
            ui.parameter('worker type').set_locked(True)

    def process_task(self, context) -> ProcessingResult:
        script = 'import os, hou\n'

        source_hip = context.param_value('hip path')
        dest_hip = source_hip
        if context.param_value('save different hip'):
            dest_hip = context.param_value('save hip path')

        if context.param_value('mask as different hip'):
            mask_path = context.param_value('mask hip path')
            script += 'def __fix_hip_env__(event_type=None):\n' \
                      '    if event_type == hou.hipFileEventType.BeforeSave:\n' \
                     f'        hou.hipFile.setName({repr(dest_hip)})\n' \
                      '        return\n' \
                      '    if event_type not in (hou.hipFileEventType.AfterSave, hou.hipFileEventType.AfterLoad) and event_type is not None:\n' \
                      '        return\n' \
                     f'    hou.hipFile.setName({repr(mask_path)})\n' \
                      'hou.hipFile.addEventCallback(__fix_hip_env__)\n'

        script += 'def __main_body__():\n'
        script += '\n'.join(f'    {line}' for line in context.param_value('script').splitlines())

        script += '\n\n' \
                 f'hou.hipFile.load({repr(source_hip)}, ignore_load_warnings=True)\n' \
                  '__main_body__()\n' \
                 f'hou.hipFile.save({repr(dest_hip)})\n'

        job = InvocationJob(['hython', ':/work_to_do.py'])
        job.set_extra_file('work_to_do.py', script)
        return ProcessingResult(job=job)

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



import random
import shlex
from fnmatch import fnmatchcase

from typing import Optional, List, Iterable


def generate_name(word_length: int = 6, max_word_length: Optional[int] = None):
    word_length = max(1, word_length)
    if max_word_length is not None:
        word_length = random.randint(word_length, max(word_length, max_word_length))

    blk1 = 'aeiouy'
    blk2 = 'bcdfghjklmnprstvw'

    if random.random() > 0.5:
        blk1, blk2 = blk2, blk1

    parts: List[str] = []
    for i in range(word_length):
        parts.append(random.choice(blk1 if i % 2 == 0 else blk2))

    parts[0] = parts[0].capitalize()
    return ''.join(parts)


def match_pattern(pattern: str, text: str) -> bool:
    """
    checks if text matches pattern, where pattern is a number of fnmatch patterns separated by whitespaces
    quotes can be used for patterns with spaces, as the whole line is treated as shell arguments, patterns are splitted using shlex.split
    so do not forget to escape quotes that are actually supposed to be quotes

    patterns can be patterns and antipatterns:
    if a pattern starts with ^ symbol - it is an antipattern.
    if you need an actual ^ symbol - you will have to double escape it - once for shlex to eat off, second time for the pattern itself

    if text matches ANY of patterns and doesn't match any antipatterns - it passes

    :param pattern:
    :param text:
    :return:
    """
    matched = False
    for subpattern in shlex.split(pattern):
        if matched and subpattern.startswith('^'):  # antipattern
            if fnmatchcase(text, subpattern[1:]):
                matched = False
        elif not matched:
            if subpattern.startswith(r'\^'):
                subpattern = subpattern[1:]
            if fnmatchcase(text, subpattern):
                matched = True
    return matched


def filter_by_pattern(pattern: str, items: Iterable[str]) -> List[str]:
    """
    takes a list of
    :param pattern: 
    :param items: 
    :return: 
    """
    return [x for x in items if match_pattern(pattern, x)]


def nice_memory_formatting(memory_bytes: int) -> str:
    prefix = ''
    if memory_bytes < 0:
        prefix = '-'
        memory_bytes *= -1
    suff = ('B', 'KB', 'MB', 'GB', 'TB', 'PB')
    next_suff = 'EB'
    for su in suff:
        if memory_bytes < 1100:
            return f'{prefix}{memory_bytes}{su}'
        memory_bytes //= 1000
    return f'{prefix}{memory_bytes}{next_suff}'



from lifeblood.basenode import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.enums import NodeParameterType
from lifeblood_alicevision_modules.base_node import AlicevisionBaseNode
from pathlib import Path
from typing import Iterable


def node_class():
    return AlicevisionFeatureMatching


class AlicevisionFeatureMatching(AlicevisionBaseNode):
    def __init__(self, name):
        super(AlicevisionFeatureMatching, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('input', 'Input Images Directory', NodeParameterType.STRING, "`task['av_camera_sfm']`")
            ui.add_parameter('featuresFolders', 'Features Dir', NodeParameterType.STRING, "`task['av_features_dir']`")  # TODO: this seem to be an array of dirs, so add support
            ui.add_parameter('imagePairsList', 'Image Pairs List', NodeParameterType.STRING, "`task['av_image_matches']`")
            ui.add_parameter('output', 'Output Feature Matching Dir', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/FeatureMatching")
            ui.add_separator()
            ui.add_parameter('describerTypes', 'Describer Types', NodeParameterType.STRING, 'sift')  # TODO: make a set of checkboxes
            ui.add_parameter('crossMatching', 'Cross Matching', NodeParameterType.BOOL, False)
            ui.add_parameter('guidedMatching', 'Guided Matching', NodeParameterType.BOOL, False)
            ui.add_parameter('matchFromKnownCameraPoses', 'Match From Known Camera Poses', NodeParameterType.BOOL, False)
            ui.add_separator()
            ui.add_parameter('distanceRatio', 'distanceRatio', NodeParameterType.FLOAT, 0.8)
            ui.add_parameter('maxIteration', 'maxIteration', NodeParameterType.INT, 2048)

            ui.add_separator()
            with ui.parameters_on_same_line_block():
                ui.add_parameter('rangeStart', 'Image Range Start', NodeParameterType.INT, 0)
                ui.add_parameter('rangeSize', 'Image Chunk Size', NodeParameterType.INT, 999999)

    @classmethod
    def label(cls) -> str:
        return 'AV Feature Matching'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'alicevision', 'feature', 'matching'

    @classmethod
    def type_name(cls) -> str:
        return 'stock_alicevision_feature_matching'

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        output = Path(context.param_value('output').strip())
        if not output.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')

        output.mkdir(parents=True, exist_ok=True)  # TODO: move this to worker

        args = ['aliceVision_featureMatching', '--verboseLevel', 'info']
        for param_name in ('input', 'featuresFolders', 'imagePairsList',
                           'describerTypes', 'crossMatching', 'guidedMatching', 'matchFromKnownCameraPoses',
                           'distanceRatio', 'maxIteration',
                           'rangeStart', 'rangeSize'):
            args += [f'--{param_name}', context.param_value(param_name)]
        args += ['--output', output]

        job = InvocationJob([str(x) for x in args])
        res = ProcessingResult(job)
        res.set_attribute('av_feature_matches', str(output))
        return res



import json

from .invocationjob import InvocationJob
from .taskspawn import TaskSpawn
from .environment_resolver import EnvironmentResolverArguments
from .processingcontext import ProcessingContext  # reexport

from typing import List, Dict, Any, Optional


class ProcessingError(RuntimeError):
    pass


class ProcessingResult:
    def __init__(self, job: Optional[InvocationJob] = None, spawn: List[TaskSpawn] = None, node_output_name: Optional[str] = None):
        self.invocation_job: Optional[InvocationJob] = job
        self.spawn_list: List[TaskSpawn] = spawn
        self.do_kill_task: bool = False
        self.attributes_to_set: Optional[Dict[str, Any]] = {}
        self.do_split_remove: bool = False
        self.split_attributes_to_set: Optional[Dict[str, Any]] = {}
        self.output_name: str = node_output_name
        self._split_attribs = None
        self._environment_resolver_arguments: Optional[EnvironmentResolverArguments] = None

    def set_node_output_name(self, newname: str):
        self.output_name = newname

    def kill_task(self):
        self.do_kill_task = True

    def remove_split(self, attributes_to_set=None):
        """
        seals this one split
        :param attributes_to_set:
        :return:
        """
        self.do_split_remove = True
        if attributes_to_set is not None:
            self.split_attributes_to_set.update(attributes_to_set)

    def set_attribute(self, key: str, value):
        self.attributes_to_set[key] = value

    def remove_attribute(self, key: str):
        self.attributes_to_set[key] = None

    def set_environment_resolver_arguments(self, args: EnvironmentResolverArguments):
        self._environment_resolver_arguments = args

    def add_spawned_task(self, spawn: TaskSpawn):
        if self.spawn_list is None:
            self.spawn_list = []
        self.spawn_list.append(spawn)

    def cancel_split_task(self):
        self._split_attribs = None

    def split_task(self, into: int):
        if into < 1:
            raise ValueError('cannot split into less than to 1 parts')

        self._split_attribs = [{} for _ in range(into)]

    def set_split_task_attrib(self, split: int, attr_name: str, attr_value):
        try:
            json.dumps(attr_value)
        except:
            raise ValueError('attribs must be json-serializable dict')
        self._split_attribs[split][attr_name] = attr_value

    def set_split_task_attribs(self, split: int, attribs: dict):
        try:
            assert isinstance(attribs, dict)
            json.dumps(attribs)
        except:
            raise ValueError('attribs must be json-serializable dict')
        self._split_attribs[split] = attribs




import os
import shutil
from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.processingcontext import ProcessingContext
from lifeblood.uidata import NodeParameterType

from typing import Iterable


def node_class():
    return FileOperations


class FileOperations(BaseNode):
    def __init__(self, name):
        super(FileOperations, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.24, 0.25, 0.48)
            ui.add_parameter('on workers', 'submit to workers', NodeParameterType.BOOL, False)
            with ui.multigroup_parameter_block('item count'):
                with ui.parameters_on_same_line_block():
                    ui.add_parameter('path', 'do', NodeParameterType.STRING, '')
                    opparam = ui.add_parameter('op', 'operation', NodeParameterType.STRING, 'nop') \
                        .add_menu((('nothing', 'nop'),
                                   ('list files', 'ls'),
                                   ('create', 'touch'),
                                   ('create dir', 'mkdir'),
                                   ('create base dir', 'mkdirbase'),
                                   ('delete', 'rm'),
                                   ('copy', 'cp'),
                                   ('move', 'mv')))
                    ui.add_parameter('other path', 'to here', NodeParameterType.STRING, '') \
                        .append_visibility_condition(opparam, 'in', ('cp', 'mv'))
                    ui.add_parameter('op res', 'save as this attribute', NodeParameterType.STRING, 'files') \
                        .append_visibility_condition(opparam, '==', 'ls')

    @classmethod
    def label(cls) -> str:
        return 'file operations'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'file', 'operation', 'create', 'delete', 'remove'

    @classmethod
    def type_name(cls) -> str:
        return 'fileops'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        item_count = context.param_value('item count')
        if context.param_value('on workers'):
            scriptlines = []  # init with no lines, so later easier to check if nothing was added to them
            for i in range(item_count):
                path = os.path.realpath(context.param_value(f'path_{i}'))
                op = context.param_value(f'op_{i}')

                if op == 'nop':
                    pass
                elif op == 'touch':
                    scriptlines.extend(
                        [f'os.makedirs(os.path.dirname({repr(path)}), exist_ok=True)',
                         f'with open({repr(path)}, "wb") as f:',
                         f'    pass'])
                elif op == 'mkdir':
                    scriptlines.append(f'os.makedirs({repr(path)}, exist_ok=True)')
                elif op == 'ls':
                    res_attr_name = context.param_value(f"op res_{i}").strip()
                    if res_attr_name == '':
                        raise ProcessingError('resulting attribute name cannot be empty')
                    scriptlines.append(f'attribs_to_set[{repr(res_attr_name)}] = os.listdir({repr(path)})')
                elif op == 'mkdirbase':
                    scriptlines.append(f'os.makedirs(os.path.dirname({repr(path)}, exist_ok=True)')
                elif op in ('cp', 'mv', 'rm'):
                    if op in ('cp', 'mv'):
                        other_path = os.path.realpath(context.param_value(f'other path_{i}'))
                        scriptlines.extend(
                            [f'if os.path.isfile({repr(path)}):',
                             f'    shutil.copy2({repr(path)}, {repr(other_path)})',
                             f'elif os.path.isdir(path):',
                             f'    shutil.copytree(path, other_path, dirs_exist_ok=True)',
                             f'else:',
                             f'    raise RuntimeError(\'cannot determine is "{path}" a file or a dir\')']
                        )
                    if op in ('mv', 'rm'):
                        scriptlines.extend(
                            [f'if os.path.exists({repr(path)}):',
                             f'    if os.path.isfile({repr(path)}):',
                             f'        os.remove({repr(path)})',
                             f'    elif os.path.isdir({repr(path)}):',
                             f'        shutil.rmtree({repr(path)})',
                             f'    else:',
                             f'        raise RuntimeError(\'cannot determine is "{path}" a file or a dir\')'])
                else:
                    raise ProcessingError(f'unknown operation: "{op}"')
            if len(scriptlines) == 0:
                return ProcessingResult()
            scriptlines = ['import os',
                           'import shutil',
                           'import lifeblood_connection',
                           'attribs_to_set = {}'
                           ] + scriptlines + \
                          ['if attribs_to_set:',
                           '    lifeblood_connection.set_attributes(attribs_to_set, blocking=True)'
                           ]

            job = InvocationJob(['python', ':/script.py'])
            job.set_extra_file('script.py', '\n'.join(scriptlines))
            return ProcessingResult(job)
        else:
            result = ProcessingResult()
            for i in range(item_count):
                path = os.path.realpath(context.param_value(f'path_{i}'))
                op = context.param_value(f'op_{i}')

                if op == 'nop':
                    pass
                elif op == 'touch':
                    os.makedirs(os.path.dirname(path), exist_ok=True)
                    with open(path, 'wb') as f:
                        pass
                elif op == 'mkdir':
                    os.makedirs(path, exist_ok=True)
                elif op == 'ls':
                    res_attr_name = context.param_value(f"op res_{i}").strip()
                    if res_attr_name == '':
                        raise ProcessingError('resulting attribute name cannot be empty')
                    result.set_attribute(res_attr_name, list(os.listdir(path)))
                elif op == 'mkdirbase':
                    os.makedirs(os.path.dirname(path), exist_ok=True)
                elif op == 'rm':
                    delete_path(path)
                elif op in ('cp', 'mv'):
                    other_path = os.path.realpath(context.param_value(f'other path_{i}'))
                    os.makedirs(os.path.dirname(other_path), exist_ok=True)
                    if os.path.isfile(path):
                        shutil.copy2(path, other_path)
                    elif os.path.isdir(path):
                        shutil.copytree(path, other_path, dirs_exist_ok=True)
                    else:
                        raise ProcessingError(f'cannot determine is "{path}" a file or a dir')
                    if op == 'mv':
                        delete_path(path)
                else:
                    raise ProcessingError(f'unknown operation: "{op}"')
            return result

    def postprocess_task(self, context: ProcessingContext) -> ProcessingResult:
        return ProcessingResult()


def delete_path(path):
    if not os.path.exists(path):
        return
    if os.path.isfile(path):
        os.remove(path)
    elif os.path.isdir(path):
        shutil.rmtree(path)
    else:
        raise ProcessingError(f'cannot determine is "{path}" a file or a dir')




from typing import Tuple


class NodeColorScheme:
    def __init__(self):
        self.__main_color = (0, 0, 0)
        self.__secondary_color = None

    def set_main_color(self, r: float, g: float, b: float):
        self.__main_color = (r, g, b)

    def set_secondary_color(self, r: float, g: float, b: float):
        self.__secondary_color = (r, g, b)

    def main_color(self):
        return self.__main_color

    def secondary_color(self):
        return self.__secondary_color


from typing import Optional, Set

class TypeMetadata:
    @property
    def label(self) -> Optional[str]:
        raise NotImplementedError()

    @property
    def tags(self) -> Set[str]:
        raise NotImplementedError()


from .messages import Message
from .message_stream import MessageSendStreamBase, MessageReceiveStreamBase
from .address import DirectAddress, AddressChain

from typing import Callable, Awaitable


class MessageReceiverFactory:
    async def create_receiver(self, address: DirectAddress, message_callback: Callable[[Message], Awaitable[bool]]) -> "MessageReceiver":
        raise NotImplementedError()


class MessageReceiver:
    def __init__(self, this_address: DirectAddress):
        self.__this_address = this_address

    def this_address(self) -> DirectAddress:
        return self.__this_address

    def stop(self):
        """
        stop running receiver async task
        """
        raise NotImplementedError()

    async def wait_till_stopped(self):
        raise NotImplementedError()


class MessageStreamFactory:
    async def open_sending_stream(self, destination: DirectAddress, source: DirectAddress) -> MessageSendStreamBase:
        raise NotImplementedError()



import sys
import os
import asyncio
import platform
import subprocess
import signal
import psutil

from .logging import get_logger

__logger = get_logger('process management')

oh_no_its_windows = platform.system() == 'Windows'

if oh_no_its_windows:
    import random
    import string
    import tempfile
    from win32job import CreateJobObject, TerminateJobObject

from typing import Optional


async def create_worker_process(args):
    """
    helper function to use to create a worker from worker pool

    :param args:
    :return:
    """
    if oh_no_its_windows:
        return await asyncio.create_subprocess_exec(*args, close_fds=True,
                                                    creationflags=subprocess.CREATE_NEW_PROCESS_GROUP)  # this is cuz win "bReAk eVeNt" can only be sent to process group)
    else:
        return await asyncio.create_subprocess_exec(*args, close_fds=True)


def send_stop_signal_to_worker(process):
    """
    helper function to tell worker process to stop
    used by worker pool

    :param process:
    :return:
    """
    if oh_no_its_windows:
        return process.send_signal(signal.CTRL_BREAK_EVENT)
    else:
        return process.send_signal(signal.SIGTERM)


async def create_process(args: list, env: dict, cwd: Optional[str] = None, stdout=subprocess.PIPE, stderr=subprocess.PIPE) -> asyncio.subprocess.Process:
    """
    helper function mainly for worker to spawn a new process with a new process group.
    NOTE: process is created with stdout, stderr set to PIPE by default! careful not to deadlock!

    all arguments are passed (almost) directly to Popen

    :param args: arguments to run
    :param env: dict of environment variables
    :param cwd: current working directory to be set for the new process
    :param stdout: same as in Popen
    :param stderr: same as in Popen
    :return:
    """
    if oh_no_its_windows:
        class _HandleWrapper:
            """
            helper class that will delete temporary file when object is GCed
            """
            def __init__(self, fd, path):
                self.__fd = fd
                self.__path = path
                self.__logger = get_logger('win32_handle_wrapper')

            def __del__(self):
                try:
                    os.close(self.__fd)
                    os.unlink(self.__path)
                except Exception:
                    self.__logger.warning(f'failed to cleanup wrapped handle to: {self.__path}')

        wrapper_code = '''
import sys
import subprocess
from win32job import AssignProcessToJobObject, CreateJobObject 
from win32api import GetCurrentProcess

job = CreateJobObject(None, {job_name})
AssignProcessToJobObject(job, GetCurrentProcess())
sys.exit(subprocess.Popen({args}, env={env}, cwd={cwd}).wait())
'''
        job_name = ''.join(random.choice(string.ascii_letters) for _ in range(64))  # we HOPE there's no name collision
        job = CreateJobObject(None, job_name)

        # wrapper code may be too big for windows's ARGMAX limitations... so we'd better save it to a file
        fd, filepath = tempfile.mkstemp('_lnchr.py')
        with open(filepath, 'w') as f:
            f.write(wrapper_code.format(job_name=repr(job_name), args=repr(args), env=repr(env), cwd=repr(cwd)))

        proc = await asyncio.create_subprocess_exec(
            sys.executable, filepath,
            stdout=stdout,
            stderr=stderr,
            cwd=cwd,
            env=None,  # we need to execute it in our env, so that win32job module is available there
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP  # this is cuz win "bReAk eVeNt" can only be sent to process group
        )
        proc._win32_job_object = job
        proc._win32_handlewrapper = _HandleWrapper(fd, filepath)  # now when process object is deleted - temp file will eventually be cleaned by gc

        return proc
    else:
        return await asyncio.create_subprocess_exec(
            *args,
            stdout=stdout,
            stderr=stderr,
            cwd=cwd,
            env=env,
            restore_signals=True,
            start_new_session=True
        )


async def kill_process_tree(process: asyncio.subprocess.Process, graceful_close_timeout=10) -> int:
    """
    kill somehow

    :param process:
    :param graceful_close_timeout:
    :return:
    """
    if oh_no_its_windows:
        return await kill_process_tree_windows(process, graceful_close_timeout)
    return await kill_process_tree_posix(process, graceful_close_timeout)


async def kill_process_tree_posix(process: asyncio.subprocess.Process, graceful_close_timeout=10) -> int:
    """
    POSIX ONLY
    this one tries to walk the process tree and SIGTERM all processes, then SIGKILL ones that are stuck
    And then SIGKILL by the process group.
    Note: process set defined by pgroup and by process tree may easily be different.

    NOTE: this ASSUMES the process being killed IS the process group leader, as created by create_process

    :param process:
    :param graceful_close_timeout:
    :return:
    """
    all_proc = []
    try:
        puproc = psutil.Process(process.pid)
        all_proc.append(puproc)
        all_proc += puproc.children(recursive=True)  # both these lines can raise NoSuchProcess
    except psutil.NoSuchProcess:
        __logger.warning(f'cannot find process with pid {process.pid}. Assuming it finished. retcode={process.returncode}')
    else:
        for proc in all_proc:
            try:
                proc.terminate()
            except psutil.NoSuchProcess:
                pass
        poll_time = 0.5
        for i in range(int(graceful_close_timeout / poll_time)):
            if not all(not proc.is_running() for proc in all_proc):
                await asyncio.sleep(poll_time)
            else:
                break
        else:
            for proc in all_proc:
                if not proc.is_running():
                    continue
                try:
                    proc.kill()
                except psutil.NoSuchProcess:
                    pass

    # just to finish off all stray processes
    try:
        __logger.debug(f'sending SIGKILL to process group {process.pid}')
        os.killpg(process.pid, signal.SIGKILL)
    except OSError as e:
        __logger.warning(f'failed to send SIGKILL to group, {str(e)}')
        pass
    return await process.wait()


async def kill_process_tree_windows(process: asyncio.subprocess.Process, graceful_close_timeout=10) -> int:
    """
    WINDOWS ONLY!

    :param process:
    :param graceful_close_timeout:
    :return:
    """
    # smth to read: https://stackoverflow.com/questions/35772001/how-to-handle-a-signal-sigint-on-a-windows-os-machine
    # first get all existing processes
    all_proc = []
    try:
        puproc = psutil.Process(process.pid)
        all_proc.append(puproc)
        all_proc += puproc.children(recursive=True)  # both these lines can raise NoSuchProcess
    except psutil.NoSuchProcess:
        __logger.warning(f'cannot find process with pid {process.pid}. Assuming it finished. retcode={process.returncode}')

    if hasattr(process, '_win32_job_object'):
        __logger.debug('terminating job object')
        TerminateJobObject(process._win32_job_object, 1)
        process._win32_job_object.close()
    else:
        __logger.warning('Failed to terminate the Job object. Trying to send BREAK event...')
        # this will actually act as SIGTERM to a pg, so here processes will have time to respond
        try:
            process.send_signal(signal.CTRL_BREAK_EVENT)
        except psutil.NoSuchProcess:
            __logger.warning('Failed to send BREAK event to process group! there might be wild processes still running!')

    # give them time to gracefully close
    poll_time = 0.5
    for i in range(int(graceful_close_timeout / poll_time)):
        if any(proc.is_running() for proc in all_proc):
            await asyncio.sleep(poll_time)
        else:
            break
    # now go through processes who did not stop and kill
    for proc in all_proc:
        try:
            proc.kill()
        except psutil.NoSuchProcess:
            pass

    return await process.wait()



import asyncio
import pickle
import os
import pathlib
import math
from copy import deepcopy
from .enums import NodeParameterType
from .processingcontext import ProcessingContext
from .node_visualization_classes import NodeColorScheme
import re

from typing import TYPE_CHECKING, TypedDict, Dict, Any, List, Set, Optional, Tuple, Union, Iterable, FrozenSet, Type, Callable

if TYPE_CHECKING:
    from .basenode import BaseNode


class ParameterExpressionError(Exception):
    def __init__(self, inner_exception):
        self.__inner_exception = inner_exception

    def __str__(self):
        return f'ParameterExpressionError: {str(self.__inner_exception)}'

    def inner_expection(self):
        return self.__inner_exception


class ParameterExpressionCastError(ParameterExpressionError):
    """
    represents error with type casting of the expression result
    """
    pass


class LayoutError(RuntimeError):
    pass


class LayoutReadonlyError(LayoutError):
    pass


# if TYPE_CHECKING:
#     class Parameter(TypedDict):
#         type: NodeParameterType
#         value: Any
class ParameterHierarchyItem:
    def __init__(self):
        self.__parent: Optional["ParameterHierarchyItem"] = None
        self.__children: Set["ParameterHierarchyItem"] = set()

    def parent(self) -> Optional["ParameterHierarchyItem"]:
        return self.__parent

    def set_parent(self, item: Optional["ParameterHierarchyItem"]):
        if self.__parent == item:
            return
        if self.__parent is not None:
            assert self in self.__parent.__children
            self.__parent._child_about_to_be_removed(self)
            self.__parent.__children.remove(self)
        self.__parent = item
        if self.__parent is not None:
            self.__parent.__children.add(self)
            self.__parent._child_added(self)

    def _child_about_to_be_removed(self, child: "ParameterHierarchyItem"):
        """
        callback for just before a child is removed
        :param child:
        :return:
        """
        pass

    def _child_added(self, child: "ParameterHierarchyItem"):
        """
        callback for just after child is added
        :param child:
        :return:
        """
        pass

    def children(self) -> FrozenSet["ParameterHierarchyItem"]:
        return frozenset(self.__children)

    def _children_definition_changed(self, children: Iterable["ParameterHierarchyItem"]):
        if self.__parent is not None:
            self.__parent._children_definition_changed([self])

    def _children_appearance_changed(self, children: Iterable["ParameterHierarchyItem"]):
        if self.__parent is not None:
            self.__parent._children_appearance_changed([self])

    def _children_value_changed(self, children: Iterable["ParameterHierarchyItem"]):
        if self.__parent is not None:
            self.__parent._children_value_changed([self])

    def visible(self) -> bool:
        return False


class ParameterHierarchyLeaf(ParameterHierarchyItem):
    def _children_definition_changed(self, children: Iterable["ParameterHierarchyItem"]):
        return

    def _children_value_changed(self, children: Iterable["ParameterHierarchyItem"]):
        return

    def _children_appearance_changed(self, children: Iterable["ParameterHierarchyItem"]):
        return

    def _child_added(self, child: "ParameterHierarchyItem"):
        raise RuntimeError('cannot add children to ParameterHierarchyLeaf')

    def _child_about_to_be_removed(self, child: "ParameterHierarchyItem"):
        raise RuntimeError('cannot remove children from ParameterHierarchyLeaf')


def evaluate_expression(expression, context: Optional[ProcessingContext]):
    try:
        return eval(expression,
                    {'os': os, 'pathlib': pathlib, 'Path': pathlib.Path, **{k: getattr(math, k) for k in dir(math) if not k.startswith('_')}},
                    context.locals() if context is not None else {})
    except Exception as e:
        raise ParameterExpressionError(e) from None


class Separator(ParameterHierarchyLeaf):
    pass


class Parameter(ParameterHierarchyLeaf):

    class DontChange:
        pass

    def __init__(self, param_name: str, param_label: Optional[str], param_type: NodeParameterType, param_val: Any, can_have_expression: bool = True, readonly: bool = False, default_value = None):
        super(Parameter, self).__init__()
        self.__name = param_name
        self.__label = param_label
        self.__type = param_type
        self.__value = None
        self.__menu_items: Optional[Dict[str, str]] = None
        self.__menu_items_order: List[str] = []
        self.__vis_when = []
        self.__force_hidden = False
        self.__is_readonly = False  # set it False until the end of constructor
        self.__locked = False  # same as readonly, but is settable by user

        self.__expression = None
        self.__can_have_expressions = can_have_expression

        self.__re_expand_pattern = re.compile(r'((?<!\\)`.*?(?<!\\)`)')  # TODO: add possibility to escape `
        self.__re_escape_backticks_pattern = re.compile(r'\\`')

        self.__hard_borders: Tuple[Optional[Union[int, float]], Optional[Union[int, float]]] = (None, None)
        self.__display_borders: Tuple[Optional[Union[int, float]], Optional[Union[int, float]]] = (None, None)

        self.__string_multiline = False
        self.__string_multiline_syntax_hint: Optional[str] = None

        # links
        self.__params_referencing_me: Set["Parameter"] = set()

        # caches
        self.__vis_cache = None

        assert default_value is None or type(default_value) == type(param_val)
        self.set_value(param_val)
        self.__default_value = default_value or param_val
        self.__is_readonly = readonly

    def name(self) -> str:
        return self.__name

    def _set_name(self, name: str):
        """
        this should only be called by layout classes
        """
        self.__name = name
        if self.parent() is not None:
            self.parent()._children_definition_changed([self])

    def label(self) -> Optional[str]:
        return self.__label

    def type(self) -> NodeParameterType:
        return self.__type

    def unexpanded_value(self, context: Optional[ProcessingContext] = None):  # TODO: why context parameter here?
        return self.__value

    def default_value(self):
        """
        note that this value will be unexpanded

        :return:
        """
        return self.__default_value

    def value(self, context: Optional[ProcessingContext] = None) -> Any:
        """
        returns value of this parameter
        :param context: optional dict like locals, for expression evaluations
        """

        if self.__expression is not None:
            result = evaluate_expression(self.__expression, context)
            # check type and cast
            try:
                if self.__type == NodeParameterType.INT:
                    result = int(result)
                elif self.__type == NodeParameterType.FLOAT:
                    result = float(result)
                elif self.__type == NodeParameterType.STRING and not isinstance(result, str):
                    result = str(result)
                elif self.__type == NodeParameterType.BOOL:
                    result = bool(result)
            except ValueError:
                raise ParameterExpressionCastError(f'could not cast {result} to {self.__type.name}') from None
            #check limits
            if self.__type in (NodeParameterType.INT, NodeParameterType.FLOAT):
                if self.__hard_borders[0] is not None and result < self.__hard_borders[0]:
                    result = self.__hard_borders[0]
                if self.__hard_borders[1] is not None and result > self.__hard_borders[1]:
                    result = self.__hard_borders[1]
            return result

        if self.__type != NodeParameterType.STRING:
            return self.__value

        # for string parameters we expand expressions in ``, kinda like bash
        parts = self.__re_expand_pattern.split(self.__value)
        for i, part in enumerate(parts):
            if part.startswith('`'):  # expression
                parts[i] = str(evaluate_expression(self.__re_escape_backticks_pattern.sub('`', part[1:-1]), context))
            else:
                parts[i] = self.__re_escape_backticks_pattern.sub('`', part)
        return ''.join(parts)
        # return self.__re_expand_pattern.sub(lambda m: str(evaluate_expression(m.group(1), context)), self.__value)

    def set_slider_visualization(self, value_min=DontChange, value_max=DontChange):  # type: (Union[int, float], Union[int, float]) -> Parameter
        """
        set a visual slider's minimum and maximum
        this does nothing to the parameter itself, and it's up to parameter renderer to interpret this data

        :return: self to be chained
        """
        if self.__type not in (NodeParameterType.INT, NodeParameterType.FLOAT):
            raise ParameterDefinitionError('cannot set limits for parameters of types other than INT and FLOAT')

        if self.__type == NodeParameterType.INT:
            value_min = int(value_min)
        elif self.__type == NodeParameterType.FLOAT:
            value_min = float(value_min)

        if self.__type == NodeParameterType.INT:
            value_max = int(value_max)
        elif self.__type == NodeParameterType.FLOAT:
            value_max = float(value_max)

        self.__display_borders = (value_min, value_max)
        return self

    def set_value_limits(self, value_min=DontChange, value_max=DontChange):  # type: (Union[int, float, None, Type[DontChange]], Union[int, float, None, Type[DontChange]]) -> Parameter
        """
        set minimum and maximum values that parameter will enforce
        None means no limit (unset limit)

        :return: self to be chained
        """
        if self.__type not in (NodeParameterType.INT, NodeParameterType.FLOAT):
            raise ParameterDefinitionError('cannot set limits for parameters of types other than INT and FLOAT')
        if value_min == self.DontChange:
            value_min = self.__hard_borders[0]
        elif value_min is not None:
            if self.__type == NodeParameterType.INT:
                value_min = int(value_min)
            elif self.__type == NodeParameterType.FLOAT:
                value_min = float(value_min)
        if value_max == self.DontChange:
            value_max = self.__hard_borders[1]
        elif value_max is not None:
            if self.__type == NodeParameterType.INT:
                value_max = int(value_max)
            elif self.__type == NodeParameterType.FLOAT:
                value_max = float(value_max)
        assert value_min != self.DontChange
        assert value_max != self.DontChange

        self.__hard_borders = (value_min, value_max)
        if value_min is not None and self.__value < value_min:
            self.__value = value_min
        if value_max is not None and self.__value > value_max:
            self.__value = value_max
        return self

    def set_text_multiline(self, syntax_hint: Optional[str] = None):
        if self.__type != NodeParameterType.STRING:
            raise ParameterDefinitionError('multiline can be only set for string parameters')
        self.__string_multiline = True
        self.__string_multiline_syntax_hint = syntax_hint
        return self

    def is_text_multiline(self):
        return self.__string_multiline

    def syntax_hint(self) -> Optional[str]:
        """
        may hint an arbitrary string hint to the renderer
        it's up to renderer to decide what to do.
        common conception is to use language name lowercase, like: python
        None means no hint
        """
        return self.__string_multiline_syntax_hint

    def display_value_limits(self) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        """
        returns a tuple of limits for display purposes.
        parameter itself ignores this totally.
        it's up to parameter renderer to interpret this info
        """
        return self.__display_borders

    def value_limits(self) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        """
        returns a tuple of hard limits.
        these limits are enforced by the parameter itself
        """
        return self.__hard_borders

    def is_readonly(self):
        return self.__is_readonly

    def is_locked(self):
        return self.__locked

    def set_locked(self, locked: bool):
        if locked == self.__locked:
            return
        self.__locked = locked
        if self.parent() is not None:
            self.parent()._children_definition_changed([self])

    def set_value(self, value: Any):
        if self.__is_readonly:
            raise ParameterReadonly()
        if self.__locked:
            raise ParameterLocked()
        if self.__type == NodeParameterType.FLOAT:
            param_value = float(value)
            if self.__hard_borders[0] is not None:
                param_value = max(param_value, self.__hard_borders[0])
            if self.__hard_borders[1] is not None:
                param_value = min(param_value, self.__hard_borders[1])
        elif self.__type == NodeParameterType.INT:
            param_value = int(value)
            if self.__hard_borders[0] is not None:
                param_value = max(param_value, self.__hard_borders[0])
            if self.__hard_borders[1] is not None:
                param_value = min(param_value, self.__hard_borders[1])
        elif self.__type == NodeParameterType.BOOL:
            param_value = bool(value)
        elif self.__type == NodeParameterType.STRING:
            param_value = str(value)
        else:
            raise NotImplementedError()
        self.__value = param_value
        for other_param in self.__params_referencing_me:
            other_param._referencing_param_value_changed(self)

        if self.parent() is not None:
            self.parent()._children_value_changed([self])

    def can_have_expressions(self):
        return self.__can_have_expressions

    def has_expression(self):
        return self.__expression is not None

    def expression(self):
        return self.__expression

    def set_expression(self, expression: Union[str, None]):
        """
        sets or removes expression from a parameter
        :param expression: either expression code or None means removing expression
        :return:
        """
        if self.__is_readonly:
            raise ParameterReadonly()
        if self.__locked:
            raise ParameterLocked()
        if not self.__can_have_expressions:
            raise ParameterCannotHaveExpressions()
        if expression != self.__expression:
            self.__expression = expression
            if self.parent() is not None:
                self.parent()._children_definition_changed([self])

    def remove_expression(self):
        self.set_expression(None)

    def _referencing_param_value_changed(self, other_parameter):
        """
        when a parameter that we are referencing changes - it will report here
        :param other_parameter:
        """
        # TODO: this now only works with referencing param in visibility condition
        # TODO: butt we want general references, including from parameter expressions
        # TODO: OOOORR will i need references for expressions at all?
        # TODO: references between node bring SOOOO much pain when serializing them separately
        if self.__vis_when:
            self.__vis_cache = None
            if self.parent() is not None and isinstance(self.parent(), ParametersLayoutBase):
                self.parent()._children_appearance_changed([self])

    def set_hidden(self, hidden):
        self.__force_hidden = hidden

    def visible(self) -> bool:
        if self.__force_hidden:
            return False
        if self.__vis_cache is not None:
            return self.__vis_cache
        if self.__vis_when:
            for other_param, op, value in self.__vis_when:
                if op == '==' and other_param.value() != value \
                        or op == '!=' and other_param.value() == value \
                        or op == '>' and other_param.value() <= value \
                        or op == '>=' and other_param.value() < value \
                        or op == '<' and other_param.value() >= value \
                        or op == '<=' and other_param.value() > value \
                        or op == 'in' and other_param.value() not in value \
                        or op == 'not in' and other_param.value() in value:
                    self.__vis_cache = False
                    return False
        self.__vis_cache = True
        return True

    def _add_referencing_me(self, other_parameter: "Parameter"):
        """
        other_parameter MUST belong to the same node to avoid cross-node references
        :param other_parameter:
        :return:
        """
        assert self.has_same_parent(other_parameter), 'references MUST belong to the same node'
        self.__params_referencing_me.add(other_parameter)

    def _remove_referencing_me(self, other_parameter: "Parameter"):
        assert other_parameter in self.__params_referencing_me
        self.__params_referencing_me.remove(other_parameter)

    def references(self) -> Tuple["Parameter", ...]:
        """
        returns tuple of parameters referenced by this parameter's definition
        static/dynamic references from expressions ARE NOT INCLUDED - they are not parameter's DEFINITION
        currently the only thing that can be a reference is parameter from visibility conditions
        """
        return tuple(x[0] for x in self.__vis_when)

    def visibility_conditions(self) -> Tuple[Tuple["Parameter", str, Union[bool, int, float, str, tuple]], ...]:
        return tuple(self.__vis_when)

    def append_visibility_condition(self, other_param: "Parameter", condition: str, value: Union[bool, int, float, str, tuple]) -> "Parameter":
        """
        condition currently can only be a simplest
        :param other_param:
        :param condition:
        :param value:
        :return: self to allow easy chaining
        """
        allowed_conditions = ('==', '!=', '>=', '<=', '<', '>', 'in', 'not in')
        if condition not in allowed_conditions:
            raise ParameterDefinitionError(f'condition must be one of: {", ".join(x for x in allowed_conditions)}')
        if condition in ('in', 'not in') and not isinstance(value, tuple):
            raise ParameterDefinitionError('for in/not in conditions value must be a tuple of possible values')
        elif condition not in ('in', 'not in') and isinstance(value, tuple):
            raise ParameterDefinitionError('value can be tuple only for in/not in conditions')

        otype = other_param.type()
        if otype == NodeParameterType.INT:
            if not isinstance(value, tuple):
                value = int(value)
        elif otype == NodeParameterType.BOOL:
            if not isinstance(value, tuple):
                value = bool(value)
        elif otype == NodeParameterType.FLOAT:
            if not isinstance(value, tuple):
                value = float(value)
        elif otype == NodeParameterType.STRING:
            if not isinstance(value, tuple):
                value = str(value)
        else:  # for future
            raise ParameterDefinitionError(f'cannot add visibility condition check based on this type of parameters: {otype}')
        self.__vis_when.append((other_param, condition, value))
        other_param._add_referencing_me(self)
        self.__vis_cache = None

        self.parent()._children_definition_changed([self])
        return self

    def add_menu(self, menu_items_pairs) -> "Parameter":
        """
        adds UI menu to parameter param_name
        :param menu_items_pairs: dict of label -> value for parameter menu. type of value MUST match type of parameter param_name. type of label MUST be string
        :return: self to allow easy chaining
        """
        # sanity check and regroup
        my_type = self.type()
        menu_items = {}
        menu_order = []
        for key, value in menu_items_pairs:
            menu_items[key] = value
            menu_order.append(key)
            if not isinstance(key, str):
                raise ParameterDefinitionError('menu label type must be string')
            if my_type == NodeParameterType.INT and not isinstance(value, int):
                raise ParameterDefinitionError(f'wrong menu value for int parameter "{self.name()}"')
            elif my_type == NodeParameterType.BOOL and not isinstance(value, bool):
                raise ParameterDefinitionError(f'wrong menu value for bool parameter "{self.name()}"')
            elif my_type == NodeParameterType.FLOAT and not isinstance(value, float):
                raise ParameterDefinitionError(f'wrong menu value for float parameter "{self.name()}"')
            elif my_type == NodeParameterType.STRING and not isinstance(value, str):
                raise ParameterDefinitionError(f'wrong menu value for string parameter "{self.name()}"')

        self.__menu_items = menu_items
        self.__menu_items_order = menu_order
        self.parent()._children_definition_changed([self])
        return self

    def has_menu(self):
        return self.__menu_items is not None

    def get_menu_items(self):
        return self.__menu_items_order, self.__menu_items

    def has_same_parent(self, other_parameter: "Parameter") -> bool:
        """
        finds if somewhere down the hierarchy there is a shared parent of self and other_parameter
        """
        my_ancestry_line = set()
        ancestor = self
        while ancestor is not None:
            my_ancestry_line.add(ancestor)
            ancestor = ancestor.parent()

        ancestor = other_parameter
        while ancestor is not None:
            if ancestor in my_ancestry_line:
                return True
            ancestor = ancestor.parent()
        return False

    def nodeui(self) -> Optional["NodeUi"]:
        """
        returns parent nodeui if it is the root of current hierarchy. otherwise returns None
        """
        ancestor = self
        while ancestor is not None:
            if isinstance(ancestor, NodeUi):
                return ancestor
            ancestor = ancestor.parent()
        return None

    def __setstate__(self, state):
        """
        overriden for easier parameter class iterations during active development.
        otherwise all node ui data should be recreated from zero in DB every time a change is made
        """
        # this init here only to init new shit when unpickling old parameters without resetting DB all the times
        self.__init__('', '', NodeParameterType.INT,  0, False)
        self.__dict__.update(state)


class ParameterError(RuntimeError):
    pass


class ParameterDefinitionError(ParameterError):
    pass


class ParameterNotFound(ParameterError):
    pass


class ParameterNameCollisionError(ParameterError):
    pass


class ParameterReadonly(ParameterError):
    pass


class ParameterLocked(ParameterError):
    pass


class ParameterCannotHaveExpressions(ParameterError):
    pass


class ParametersLayoutBase(ParameterHierarchyItem):
    def __init__(self):
        super(ParametersLayoutBase, self).__init__()
        self.__parameters: Dict[str, Parameter] = {}  # just for quicker access
        self.__layouts: Set[ParametersLayoutBase] = set()
        self.__block_ui_callbacks = False

    def initializing_interface_lock(self):
        class _iiLock:
            def __init__(self, lockable):
                self.__nui = lockable
                self.__prev_state = False

            def __enter__(self):
                self.__prev_state = self.__nui._ParametersLayoutBase__block_ui_callbacks
                self.__nui._ParametersLayoutBase__block_ui_callbacks = True

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.__nui._ParametersLayoutBase__block_ui_callbacks = self.__prev_state

        return _iiLock(self)

    def _is_initialize_lock_set(self):
        return self.__block_ui_callbacks

    def add_parameter(self, new_parameter: Parameter):
        self.add_generic_leaf(new_parameter)

    def add_generic_leaf(self, item: ParameterHierarchyLeaf):
        if not self._is_initialize_lock_set():
            raise LayoutError('initializing interface not inside initializing_interface_lock')
        item.set_parent(self)

    def add_layout(self, new_layout: "ParametersLayoutBase"):
        if not self._is_initialize_lock_set():
            raise LayoutError('initializing interface not inside initializing_interface_lock')
        new_layout.set_parent(self)

    def items(self, recursive=False) -> Iterable["ParameterHierarchyItem"]:
        for child in self.children():
            yield child
            if not recursive:
                continue
            elif isinstance(child, ParametersLayoutBase):
                for child_param in child.parameters(recursive=recursive):
                    yield child_param

    def parameters(self, recursive=False) -> Iterable[Parameter]:
        for item in self.items(recursive=recursive):
            if isinstance(item, Parameter):
                yield item

    def parameter(self, name: str) -> Parameter:
        if name in self.__parameters:
            return self.__parameters[name]
        for layout in self.__layouts:
            try:
                return layout.parameter(name)
            except ParameterNotFound:
                continue
        raise ParameterNotFound(f'parameter {name} not found in layout hierarchy')

    def visible(self) -> bool:
        return len(self.children()) != 0 and any(x.visible() for x in self.items())

    def _child_added(self, child: "ParameterHierarchyItem"):
        super(ParametersLayoutBase, self)._child_added(child)
        if isinstance(child, Parameter):
            # check global parameter name uniqueness
            rootparent = self
            while isinstance(rootparent.parent(), ParametersLayoutBase):
                rootparent = rootparent.parent()
            if child.name() in (x.name() for x in rootparent.parameters(recursive=True) if x != child):
                raise ParameterNameCollisionError('cannot add parameters with the same name to the same layout hierarchy')
            self.__parameters[child.name()] = child
        elif isinstance(child, ParametersLayoutBase):
            self.__layouts.add(child)
            # check global parameter name uniqueness
            rootparent = self
            while isinstance(rootparent.parent(), ParametersLayoutBase):
                rootparent = rootparent.parent()
            new_params = list(child.parameters(recursive=True))
            existing_params = set(x.name() for x in rootparent.parameters(recursive=True) if x not in new_params)
            for new_param in new_params:
                if new_param.name() in existing_params:
                    raise ParameterNameCollisionError('cannot add parameters with the same name to the same layout hierarchy')

    def _child_about_to_be_removed(self, child: "ParameterHierarchyItem"):
        if isinstance(child, Parameter):
            del self.__parameters[child.name()]
        elif isinstance(child, ParametersLayoutBase):
            self.__layouts.remove(child)
        super(ParametersLayoutBase, self)._child_about_to_be_removed(child)

    def _children_definition_changed(self, changed_children: Iterable["ParameterHierarchyItem"]):
        """
        :param children:
        :return:
        """
        super(ParametersLayoutBase, self)._children_definition_changed(changed_children)
        # check self.__parameters consistency
        reversed_parameters: Dict[Parameter, str] = {v: k for k, v in self.__parameters.items()}
        for child in changed_children:
            if not isinstance(child, Parameter):
                continue
            if child in reversed_parameters:
                del self.__parameters[reversed_parameters[child]]
                self.__parameters[child.name()] = child

    def _children_value_changed(self, children: Iterable["ParameterHierarchyItem"]):
        """
        :param children:
        :return:
        """
        super(ParametersLayoutBase, self)._children_value_changed(children)
        
    def _children_appearance_changed(self, children: Iterable["ParameterHierarchyItem"]):
        super(ParametersLayoutBase, self)._children_appearance_changed(children)

    def relative_size_for_child(self, child: ParameterHierarchyItem) -> Tuple[float, float]:
        """
        get relative size of a child in this layout
        the exact interpretation of size is up to subclass to decide
        :param child:
        :return:
        """
        raise NotImplementedError()


class OrderedParametersLayout(ParametersLayoutBase):
    def __init__(self):
        super(OrderedParametersLayout, self).__init__()
        self.__parameter_order: List[ParameterHierarchyItem] = []

    def _child_added(self, child: "ParameterHierarchyItem"):
        super(OrderedParametersLayout, self)._child_added(child)
        self.__parameter_order.append(child)
        
    def _child_about_to_be_removed(self, child: "ParameterHierarchyItem"):
        self.__parameter_order.remove(child)
        super(OrderedParametersLayout, self)._child_about_to_be_removed(child)

    def items(self, recursive=False):
        """
        unlike base method, we need to return parameters in order
        :param recursive:
        :return:
        """
        for child in self.__parameter_order:
            yield child
            if not recursive:
                continue
            elif isinstance(child, ParametersLayoutBase):
                for child_param in child.items(recursive=recursive):
                    yield child_param

    def relative_size_for_child(self, child: ParameterHierarchyItem) -> Tuple[float, float]:
        """
        get relative size of a child in this layout
        the exact interpretation of size is up to subclass to decide
        :param child:
        :return:
        """
        assert child in self.children()
        return 1.0, 1.0


class VerticalParametersLayout(OrderedParametersLayout):
    """
    simple vertical parameter layout.
    """
    pass


class CollapsableVerticalGroup(VerticalParametersLayout):
    """
    a vertical parameter layout to be drawn as collapsable block
    """
    def __init__(self, group_name, group_label):
        super(CollapsableVerticalGroup, self).__init__()

        # for now it's here just to ensure name uniqueness. in future - maybe store collapsed state
        self.__unused_param = Parameter(group_name, group_name, NodeParameterType.BOOL, True)

        self.__group_name = group_name
        self.__group_label = group_label

    def is_collapsed(self):
        return True

    def name(self):
        return self.__group_name

    def label(self):
        return self.__group_label


class OneLineParametersLayout(OrderedParametersLayout):
    """
    horizontal parameter layout.
    unlike vertical, this one has to keep track of portions of line it's parameters are taking
    parameters of this group should be rendered in one line
    """
    def __init__(self):
        super(OneLineParametersLayout, self).__init__()
        self.__hsizes = {}

    def _children_appearance_changed(self, children: Iterable["ParameterHierarchyItem"]):
        super(ParametersLayoutBase, self)._children_appearance_changed(children)
        self.__hsizes = {}

    def _children_definition_changed(self, children: Iterable["ParameterHierarchyItem"]):
        super(OneLineParametersLayout, self)._children_definition_changed(children)
        self.__hsizes = {}

    def relative_size_for_child(self, child: ParameterHierarchyItem) -> Tuple[float, float]:
        assert child in self.children()
        if child not in self.__hsizes:
            self._update_hsizes()
        assert child in self.__hsizes
        return self.__hsizes[child], 1.0

    def _update_hsizes(self):
        self.__hsizes = {}
        totalitems = 0
        for item in self.items():
            if item.visible():
                totalitems += 1
        if totalitems == 0:
            uniform_size = 1.0
        else:
            uniform_size = 1.0 / float(totalitems)
        for item in self.items():
            self.__hsizes[item] = uniform_size


class MultiGroupLayout(OrderedParametersLayout):
    """
    this group can dynamically spawn more parameters according to it's template
    spawning more parameters does NOT count as definition change
    """
    def __init__(self, name, label=None):
        super(MultiGroupLayout, self).__init__()
        self.__template: Union[ParametersLayoutBase, Parameter, None] = None
        if label is None:
            label = 'count'
        self.__count_param = Parameter(name, label, NodeParameterType.INT, 0, can_have_expression=False)
        self.__count_param.set_parent(self)
        self.__last_count = 0
        self.__nested_indices = []

    def nested_indices(self):
        """
        if a multiparam is inside other multiparams - those multiparams should add their indices
        to this one, so that this multiparam will be able to uniquely and predictable name it's parameters
        """
        return tuple(self.__nested_indices)

    def __append_nested_index(self, index: int):
        """
        this should be called only when a multiparam is instanced by another multiparam
        """
        self.__nested_indices.append(index)

    def set_spawning_template(self, layout: ParametersLayoutBase):
        self.__template = deepcopy(layout)

    def add_layout(self, new_layout: "ParametersLayoutBase"):
        """
        this function is unavailable cuz of the nature of this layout
        """
        raise LayoutError('NO')

    def add_parameter(self, new_parameter: Parameter):
        """
        this function is unavailable cuz of the nature of this layout
        """
        raise LayoutError('NO')

    def add_template_instance(self):
        self.__count_param.set_value(self.__count_param.value() + 1)

    def _children_value_changed(self, children: Iterable["ParameterHierarchyItem"]):

        for child in children:
            if child == self.__count_param:
                break
        else:
            super(MultiGroupLayout, self)._children_value_changed(children)
            return
        if self.__count_param.value() < 0:
            self.__count_param.set_value(0)
            super(MultiGroupLayout, self)._children_value_changed(children)
            return

        new_count = self.__count_param.value()
        if self.__last_count < new_count:
            if self.__template is None:
                raise LayoutError('template is not set')
            for _ in range(new_count - self.__last_count):
                # note: the check below is good, but it's not needed currently, cuz visibility condition on append checks common parent
                # and nodes not from template do not share parents with template, so that prevents external references
                for param in self.__template.parameters(recursive=True):
                    # sanity check - for now we only support references within the same template block only
                    for ref_param in param.references():
                        if not ref_param.has_same_parent(param):
                            raise ParameterDefinitionError('Parameters within MultiGroupLayout\'s template currently cannot reference outer parameters')
                ##
                new_layout = deepcopy(self.__template)
                i = len(self.children()) - 1
                for param in new_layout.parameters(recursive=True):
                    param._set_name(param.name() + '_' + '.'.join(str(x) for x in (*self.nested_indices(), i)))
                    parent = param.parent()
                    if isinstance(parent, MultiGroupLayout):
                        for idx in self.nested_indices():
                            parent.__append_nested_index(idx)
                        parent.__append_nested_index(i)
                new_layout.set_parent(self)
        elif self.__last_count > self.__count_param.value():
            for _ in range(self.__last_count - new_count):
                instances = list(self.items(recursive=False))
                assert len(instances) > 1
                instances[-1].set_parent(None)
        self.__last_count = new_count
        super(MultiGroupLayout, self)._children_value_changed(children)
        
    def _child_added(self, child: "ParameterHierarchyItem"):
        super(MultiGroupLayout, self)._child_added(child)
    
    def _child_about_to_be_removed(self, child: "ParameterHierarchyItem"):
        super(MultiGroupLayout, self)._child_about_to_be_removed(child)


class _SpecialOutputCountChangingLayout(VerticalParametersLayout):
    def __init__(self, nodeui: "NodeUi", parameter_name, parameter_label):
        super(_SpecialOutputCountChangingLayout, self).__init__()
        self.__my_nodeui = nodeui
        newparam = Parameter(parameter_name, parameter_label, NodeParameterType.INT, 2, can_have_expression=False)
        newparam.set_value_limits(2)
        with self.initializing_interface_lock():
            self.add_parameter(newparam)

    def add_layout(self, new_layout: "ParametersLayoutBase"):
        """
        this function is unavailable cuz of the nature of this layout
        """
        raise LayoutError('NO')

    def add_parameter(self, new_parameter: Parameter):
        """
        this function is unavailable cuz of the nature of this layout
        """
        if len(list(self.parameters())) > 0:
            raise LayoutError('NO')
        super(_SpecialOutputCountChangingLayout, self).add_parameter(new_parameter)

    def _children_value_changed(self, children: Iterable["ParameterHierarchyItem"]):
        # we expect this special layout to have only one single specific child
        child = None
        for child in children:
            break
        if child is None:
            return
        assert isinstance(child, Parameter)
        new_num_outputs = child.value()
        num_outputs = len(self.__my_nodeui.outputs_names())
        if num_outputs == new_num_outputs:
            return

        if num_outputs < new_num_outputs:
            for i in range(num_outputs, new_num_outputs):
                self.__my_nodeui._add_output_unsafe(f'output{i}')
        else:  # num_outputs > new_num_outputs
            for _ in range(new_num_outputs, num_outputs):
                self.__my_nodeui._remove_last_output_unsafe()
        self.__my_nodeui._outputs_definition_changed()


class NodeUiError(RuntimeError):
    pass


class NodeUiDefinitionError(RuntimeError):
    pass


class NodeUi(ParameterHierarchyItem):
    def __init__(self, attached_node: "BaseNode"):
        super(NodeUi, self).__init__()
        self.__parameter_layout = VerticalParametersLayout()
        self.__parameter_layout.set_parent(self)
        self.__attached_node: Optional[BaseNode] = attached_node
        self.__block_ui_callbacks = False
        self.__lock_ui_readonly = False
        self.__postpone_ui_callbacks = False
        self.__postponed_callbacks = None
        self.__inputs_names = ('main',)
        self.__outputs_names = ('main',)

        self.__groups_stack = []

        self.__have_output_parameter_set: bool = False

        # default colorscheme
        self.__color_scheme = NodeColorScheme()
        self.__color_scheme.set_main_color(0.1882, 0.2510, 0.1882)  # dark-greenish

    def is_attached_to_node(self):
        return self.__attached_node is not None

    def attach_to_node(self, node: "BaseNode"):
        self.__attached_node = node

    def color_scheme(self):
        return self.__color_scheme

    def main_parameter_layout(self):
        return self.__parameter_layout

    def parent(self) -> Optional["ParameterHierarchyItem"]:
        return None

    def set_parent(self, item: Optional["ParameterHierarchyItem"]):
        if item is not None:
            raise RuntimeError('NodeUi class is supposed to be tree root')

    def initializing_interface_lock(self):
        class _iiLock:
            def __init__(self, lockable):
                self.__nui = lockable
                self.__prev_state = None

            def __enter__(self):
                self.__prev_state = self.__nui._NodeUi__block_ui_callbacks
                self.__nui._NodeUi__block_ui_callbacks = True

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.__nui._NodeUi__block_ui_callbacks = self.__prev_state

        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        return _iiLock(self)

    def lock_interface_readonly(self):
        raise NotImplementedError("read trello task, read TODO. this do NOT work multitheaded, leads to permalocks, needs rethinking")
        class _roLock:
            def __init__(self, lockable):
                self.__nui = lockable
                self.__prev_state = None

            def __enter__(self):
                self.__prev_state = self.__nui._NodeUi__lock_ui_readonly
                self.__nui._NodeUi__lock_ui_readonly = True

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.__nui._NodeUi__lock_ui_readonly = self.__prev_state

        return _roLock(self)

    def postpone_ui_callbacks(self):
        """
        use this in with-statement
        for mass change of parameters it may be more efficient to perform changes in batches
        """
        class _iiPostpone:
            def __init__(self, nodeui):
                self.__nui = nodeui
                self.__val = None

            def __enter__(self):
                if not self.__nui._NodeUi__postpone_ui_callbacks:
                    assert self.__nui._NodeUi__postponed_callbacks is None
                    self.__val = self.__nui._NodeUi__postpone_ui_callbacks
                    self.__nui._NodeUi__postpone_ui_callbacks = True
                # otherwise: already blocked - we are in nested block, ignore

            def __exit__(self, exc_type, exc_val, exc_tb):
                if self.__val is None:
                    return
                assert not self.__val  # nested block should do nothing
                self.__nui._NodeUi__postpone_ui_callbacks = self.__val
                if self.__nui._NodeUi__postponed_callbacks is not None:
                    self.__nui._NodeUi__ui_callback(self.__nui._NodeUi__postponed_callbacks)
                    self.__nui._NodeUi__postponed_callbacks = None

        return _iiPostpone(self)

    class _slwrapper:
        def __init__(self, ui: "NodeUi", layout_creator, layout_creator_kwargs=None):
            self.__ui = ui
            self.__layout_creator = layout_creator
            self.__layout_creator_kwargs = layout_creator_kwargs or {}

        def __enter__(self):
            new_layout = self.__layout_creator(**self.__layout_creator_kwargs)
            self.__ui._NodeUi__groups_stack.append(new_layout)
            with self.__ui._NodeUi__parameter_layout.initializing_interface_lock():
                self.__ui._NodeUi__parameter_layout.add_layout(new_layout)

        def __exit__(self, exc_type, exc_val, exc_tb):
            layout = self.__ui._NodeUi__groups_stack.pop()
            self.__ui._add_layout(layout)

    def parameters_on_same_line_block(self):
        """
        use it in with statement
        :return:
        """
        return self.parameter_layout_block(OneLineParametersLayout)

    def parameter_layout_block(self, parameter_layout_producer: Callable[[], ParametersLayoutBase]):
        """
        arbitrary simple parameter override block
        use it in with statement
        :return:
        """
        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        return NodeUi._slwrapper(self, parameter_layout_producer)

    def add_parameter_to_control_output_count(self, parameter_name: str, parameter_label: str):
        """
        a very special function for a very special case when you want the number of outputs to be controlled
        by a parameter

        from now on output names will be: 'main', 'output1', 'output2', ...

        :return:
        """
        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        if self.__have_output_parameter_set:
            raise NodeUiDefinitionError('there can only be one parameter to control output count')
        self.__have_output_parameter_set = True
        self.__outputs_names = ('main', 'output1')

        with self.parameter_layout_block(lambda: _SpecialOutputCountChangingLayout(self, parameter_name, parameter_label)):
            # no need to do anything, with block will add that layout to stack, and parameter is created in that layout's constructor
            layout = self.current_layout()
            # this layout should always have exactly one parameter
            assert len(list(layout.parameters())) == 1, f'oh no, {len(list(layout.parameters()))}'
        return layout.parameter(parameter_name)

    def multigroup_parameter_block(self, name: str, label: Optional[str] = None):
        """
        use it in with statement
        creates a block like multiparameter block in houdini
        any parameters added will be actually added to template to be instanced later as needed
        :return:
        """
        class _slwrapper_multi:
            def __init__(self, ui: "NodeUi", name: str, label: Optional[str] = None):
                self.__ui = ui
                self.__new_layout = None
                self.__name = name
                self.__label = label

            def __enter__(self):
                self.__new_layout = VerticalParametersLayout()
                self.__ui._NodeUi__groups_stack.append(self.__new_layout)
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                assert self.__ui._NodeUi__groups_stack.pop() == self.__new_layout
                with self.__ui._NodeUi__parameter_layout.initializing_interface_lock():
                    multi_layout = MultiGroupLayout(self.__name, self.__label)
                    with multi_layout.initializing_interface_lock():
                        multi_layout.set_spawning_template(self.__new_layout)
                    self.__ui._add_layout(multi_layout)

            def multigroup(self) -> VerticalParametersLayout:
                return self.__new_layout

        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        return _slwrapper_multi(self, name, label)

    def current_layout(self):
        """
        get current layout to which add_parameter would add parameter
        this can be main nodeUI's layout, but can be something else, if we are in some with block,
        like for ex: collapsable_group_block or parameters_on_same_line_block

        :return:
        """
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        layout = self.__parameter_layout
        if len(self.__groups_stack) != 0:
            layout = self.__groups_stack[-1]
        return layout

    def collapsable_group_block(self, group_name: str, group_label: str = ''):
        """
        use it in with statement
        creates a visually distinct group of parameters that renderer should draw as a collapsable block

        :return:
        """
        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        return NodeUi._slwrapper(self, CollapsableVerticalGroup, {'group_name': group_name, 'group_label': group_label})

    def _add_layout(self, new_layout):
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        layout = self.__parameter_layout
        if len(self.__groups_stack) != 0:
            layout = self.__groups_stack[-1]
        with layout.initializing_interface_lock():
            layout.add_layout(new_layout)

    def add_parameter(self, param_name: str, param_label: Optional[str], param_type: NodeParameterType, param_val: Any, can_have_expressions: bool = True, readonly: bool = False) -> Parameter:
        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        layout = self.__parameter_layout
        if len(self.__groups_stack) != 0:
            layout = self.__groups_stack[-1]
        with layout.initializing_interface_lock():
            newparam = Parameter(param_name, param_label, param_type, param_val, can_have_expressions, readonly)
            layout.add_parameter(newparam)
        return newparam

    def add_separator(self):
        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        layout = self.__parameter_layout
        if len(self.__groups_stack) != 0:
            layout = self.__groups_stack[-1]
        with layout.initializing_interface_lock():
            newsep = Separator()
            layout.add_generic_leaf(newsep)
        return newsep

    def add_input(self, input_name):
        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        if input_name not in self.__inputs_names:
            self.__inputs_names += (input_name,)

    def _add_output_unsafe(self, output_name):
        if output_name not in self.__outputs_names:
            self.__outputs_names += (output_name,)

    def add_output(self, output_name):
        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        if self.__have_output_parameter_set:
            raise NodeUiDefinitionError('cannot add outputs when output count is controlled by a parameter')
        return self._add_output_unsafe(output_name)

    def _remove_last_output_unsafe(self):
        if len(self.__outputs_names) < 2:
            return
        self.__outputs_names = self.__outputs_names[:-1]

    def remove_last_output(self):
        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        if not self.__block_ui_callbacks:
            raise NodeUiDefinitionError('initializing NodeUi interface not inside initializing_interface_lock')
        if self.__have_output_parameter_set:
            raise NodeUiDefinitionError('cannot add outputs when output count is controlled by a parameter')
        return self._remove_last_output_unsafe()

    def add_output_for_spawned_tasks(self):
        return self.add_output('spawned')

    def _children_definition_changed(self, children: Iterable["ParameterHierarchyItem"]):
        self.__ui_callback(definition_changed=True)

    def _children_value_changed(self, children: Iterable["ParameterHierarchyItem"]):
        self.__ui_callback(definition_changed=False)

    def _outputs_definition_changed(self):  #TODO: not entirely sure how safe this is right now
        self.__ui_callback(definition_changed=True)

    def __ui_callback(self, definition_changed=False):
        if self.__lock_ui_readonly:
            raise LayoutReadonlyError()
        if self.__postpone_ui_callbacks:
            # so we save definition_changed to __postponed_callbacks
            self.__postponed_callbacks = self.__postponed_callbacks or definition_changed
            return

        if self.__attached_node is not None and not self.__block_ui_callbacks:
            self.__attached_node._ui_changed(definition_changed)

    def inputs_names(self) -> Tuple[str]:
        return self.__inputs_names

    def outputs_names(self) -> Tuple[str]:
        return self.__outputs_names

    def parameter(self, param_name: str) -> Parameter:
        return self.__parameter_layout.parameter(param_name)

    def parameters(self) -> Iterable[Parameter]:
        return self.__parameter_layout.parameters(recursive=True)

    def items(self, recursive=False) -> Iterable[ParameterHierarchyItem]:
        return self.__parameter_layout.items(recursive=recursive)

    def __deepcopy__(self, memo):
        cls = self.__class__
        crap = cls.__new__(cls)
        newdict = self.__dict__.copy()
        newdict['_NodeUi__attached_node'] = None
        newdict['_NodeUi__lock_ui_readonly'] = False
        assert id(self) not in memo
        memo[id(self)] = crap  # to avoid recursion, though manual tells us to treat memo as opaque object
        for k, v in newdict.items():
            crap.__dict__[k] = deepcopy(v, memo)
        return crap

    def __setstate__(self, state):
        ensure_attribs = {  # this exists only for the ease of upgrading NodeUi classes during development
            '_NodeUi__lock_ui_readonly': False,
            '_NodeUi__postpone_ui_callbacks': False
        }
        self.__dict__.update(state)
        for attrname, default_value in ensure_attribs.items():
            if not hasattr(self, attrname):
                setattr(self, attrname, default_value)

    def serialize(self) -> bytes:
        """
        note - this serialization disconnects the node to which this UI is connected
        :return:
        """
        obj = deepcopy(self)
        assert obj.__attached_node is None
        return pickle.dumps(obj)

    async def serialize_async(self) -> bytes:
        return await asyncio.get_event_loop().run_in_executor(None, self.serialize)

    def __repr__(self):
        return 'NodeUi: ' + ', '.join(('%s: %s' % (x.name() if isinstance(x, Parameter) else '-layout-', x) for x in self.__parameter_layout.items()))

    @classmethod
    def deserialize(cls, data: bytes) -> "NodeUi":
        return pickle.loads(data)

    @classmethod
    async def deserialize_async(cls, data: bytes) -> "NodeUi":
        return await asyncio.get_event_loop().run_in_executor(None, cls.deserialize, data)



import asyncio


class ComponentBase:
    def __init__(self):
        super().__init__()
        self.__stop_event = asyncio.Event()
        self.__main_task = None
        self.__main_task_is_ready = asyncio.Event()

    @property
    def _stop_event(self):
        return self.__stop_event

    def _main_task_is_ready_now(self):
        """
        subclass needs to call this once starting initializations are done
        to signal that component is in running state
        """
        self.__main_task_is_ready.set()

    async def start(self):
        if self.__main_task is not None:
            raise RuntimeError('already started')
        self.__main_task = asyncio.create_task(self._main_task())
        ready_waiter = asyncio.create_task(self.__main_task_is_ready.wait())
        done, others = await asyncio.wait([ready_waiter, self.__main_task], return_when=asyncio.FIRST_COMPLETED)
        if self.__main_task in done:  # means it raised an error
            for other in others:
                other.cancel()
            await self.__main_task

    def stop(self):
        if self.__main_task is None:
            raise RuntimeError('not started')
        self.__stop_event.set()

    async def wait_till_stops(self):
        await self.__stop_event.wait()
        return await self.__main_task

    def _main_task(self):
        """
        should return the coroutine that will produce the main task to run by the component
        """
        raise NotImplementedError('override this with the main task')



import os
import sys
from pathlib import Path
from typing import Optional

org = 'xxx'
basename = 'lifeblood'

env_var_name = 'LIFEBLOOD_CONFIG_LOCATION'


def config_path(config_name: str, subname: Optional[str] = None) -> Path:
    return config_unexpanded_path(config_name, subname).expanduser()


def config_unexpanded_path(config_name: str, subname: Optional[str] = None) -> Path:
    if env_var_name in os.environ:
        return Path(os.environ[env_var_name])/subname/config_name
    base = Path('~')
    if subname is None:
        subname = 'common'
    if '.' in subname:
        subname = Path(*subname.split('.'))
    if sys.platform.startswith('linux'):
        return base/basename/subname/config_name
    if sys.platform.startswith('win'):
        return base/basename/subname/config_name
    elif sys.platform.startswith('darwin'):
        return base/'Library'/'Preferences'/basename/subname/config_name
    else:
        raise NotImplementedError(f'da heck is {sys.platform} anyway??')
    # if sys.platform.startswith('linux'):
    #     return base/'.local'/'share'/org/subname/config_name
    # if sys.platform.startswith('win'):
    #     return base/'AppData'/'Roaming'/org/subname/config_name
    # elif sys.platform.startswith('darwin'):
    #     return base/'Library'/'Application Support'/org/subname/config_name


def log_path(log_name: Optional[str], subname: Optional[str] = None, ensure_path_exists=True) -> Optional[Path]:
    path = log_unexpanded_path(log_name, subname).expanduser()
    if not ensure_path_exists:
        return path
    if not path.exists():
        try:
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch(exist_ok=True)
        except Exception:
            return None
    if not os.access(path, os.W_OK):
        return None
    return path


def log_unexpanded_path(log_name: Optional[str], subname: Optional[str] = None) -> Path:
    if sys.platform.startswith('linux'):
        log_path = Path(os.environ.get("XDG_DATA_HOME", "~/.local/share"))
    elif sys.platform.startswith("win"):
        log_path = Path(os.environ["LOCALAPPDATA"])
    elif sys.platform.startswith("darwin"):
        log_path = Path("~/Library/Application Support")
    else:
        raise NotImplementedError(f'da heck is {sys.platform} anyway??')
    log_path /= basename
    if subname:
        log_path /= subname
    if log_name:
        log_path /= log_name
    return log_path


def default_main_database_location() -> Path:
    return config_unexpanded_path('main.db', 'scheduler')



from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment
from lifeblood.text import filter_by_pattern

from typing import Iterable, Optional


class RopBaseNode(BaseNodeWithTaskRequirements):
    def __init__(self, name):
        super(RopBaseNode, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_output_for_spawned_tasks()
            ui.add_parameter('hip path', 'hip file path', NodeParameterType.STRING, "`task['hipfile']`")
            with ui.parameters_on_same_line_block():
                mask_hip = ui.add_parameter('mask as different hip', 'mask as different hip file', NodeParameterType.BOOL, False)
                ui.add_parameter('mask hip path', '', NodeParameterType.STRING, "`task.get('hipfile_orig', task['hipfile'])`")\
                    .append_visibility_condition(mask_hip, '==', True)
            ui.add_parameter('driver path', 'rop node path', NodeParameterType.STRING, "`task['hipdriver']`")
            ui.add_parameter('attrs to context', 'set these attribs as context', NodeParameterType.STRING, '')
            ui.add_parameter('scene file output', 'scene description file path', NodeParameterType.STRING, "")
            with ui.parameters_on_same_line_block():
                skipparam = ui.add_parameter('skip if exists', 'skip if result already exists', NodeParameterType.BOOL, False)
                ui.add_parameter('gen for skipped', 'generate children for skipped', NodeParameterType.BOOL, True).append_visibility_condition(skipparam, '==', True)
            ui.add_parameter('attrs', 'attributes to copy to children', NodeParameterType.STRING, '')

            ui.parameter('worker type').set_hidden(True)
            ui.parameter('worker type').set_locked(True)

    def _take_parm_name(self, context) -> str:
        raise NotImplementedError()

    def _parms_to_set_before_render(self, context) -> dict:
        return {}

    def _scene_file_getting_code(self, context) -> Optional[str]:
        """
        if string is returned - it's expected to be body of a function returning str
        aruments provided to the function: node - driver node, frame - frame currently being rendered
        code is expected to return string (scene_file path)
        result of _scene_file_parm_name is ignored then

        :param context:
        :return:
        """
        return None

    def _scene_file_parm_name(self, context) -> str:
        raise NotImplementedError()

    def _node_specialization_code(self, context) -> Optional[str]:
        """
        if string is returned - it's expected to be body of a function returning str
        arguments provided to the function: node - driver node
        code is expected to return node

        sometimes you want flexibility of allowing providing multiple types of driver nodes in the interface
        in that ces this function code must find the needed node from node provided

        :param context:
        :return:
        """
        return None

    def _image_getting_code(self, context) -> Optional[str]:
        """
        if string is returned - it's expected to be body of a function returning str
        aruments provided to the function: node - driver node, frame - frame currently being rendered
        code is expected to return string (image path)
        result of _image_path_parm_name is ignored then

        :param context:
        :return:
        """
        return None

    def _image_path_parm_name(self, context) -> str:
        raise NotImplementedError()

    def process_task(self, context) -> ProcessingResult:
        """
        this node expects to find the following attributes:
        frames
        hipfile
        hipdriver

        :param context:
        :return:
        """
        take_parm_name = self._take_parm_name(context)
        parms_to_set_before_render = self._parms_to_set_before_render(context)
        scene_file_code = self._scene_file_getting_code(context)
        scene_file_parm_name = self._scene_file_parm_name(context)
        image_path_code = self._image_getting_code(context)
        image_path_parm_name = self._image_path_parm_name(context)
        node_specialization_code = self._node_specialization_code(context)

        attrs = context.task_attributes()
        if any(x not in attrs for x in ('frames',)):
            raise ProcessingError('required attribute "frames" not found')
        hippath = context.param_value('hip path')
        driverpath = context.param_value('driver path')
        frames = attrs['frames']
        matching_attrnames = filter_by_pattern(context.param_value('attrs'), attrs.keys())
        attr_to_trans = tuple((x, attrs[x]) for x in matching_attrnames if x not in ('frames', 'file'))
        attr_to_context = filter_by_pattern(context.param_value('attrs to context'), attrs.keys())

        env = InvocationEnvironment()

        spawnlines = \
            f"        filepath = _that_one_scene_file_path_getting_function(node, frame)\n" \
            f"        outimage = _that_one_image_path_getting_function(node, frame)\n" \
            f"        attrs = {{'frames': [frame], 'file': filepath, 'hipfile': {repr(hippath)}, 'outimage': outimage}}\n" \
            f"        for attr, val in {repr(attr_to_trans)}:\n" \
            f"            attrs[attr] = val\n" \
            f"        lifeblood_connection.create_task(node.name() + '_spawned frame %g' % frame, attrs)\n"

        if not self.is_output_connected('spawned'):
            spawnlines = ''

        script = \
            f'import os\n' \
            f'import hou\n' \
            f'import lifeblood_connection\n'

        script += 'def _that_one_image_path_getting_function(node, frame) -> str:\n'
        if image_path_code:
            script += '\n'.join(f'    {x}' for x in image_path_code.splitlines()) + '\n\n'
        else:
            script += f'    return node.evalParm({repr(image_path_parm_name)})\n\n'

        script += 'def _that_one_scene_file_path_getting_function(node, frame) -> str:\n'
        if scene_file_code:
            script += '\n'.join(f'    {x}' for x in scene_file_code.splitlines()) + '\n\n'
        else:
            script += f'    return node.evalParm({repr(scene_file_parm_name)})\n\n'

        if node_specialization_code:
            script += 'def _that_one_node_specialization_function(node):\n'
            script += '\n'.join(f'    {x}' for x in node_specialization_code.splitlines()) + '\n\n'

        if context.param_value('mask as different hip'):
            mask_path = context.param_value('mask hip path')
            script += 'def __fix_hip_env__(event_type=None):\n' \
                      '    if event_type == hou.hipFileEventType.BeforeSave:\n' \
                      '        hou.hipFile.setName(os.devnull)\n' \
                      '        return\n' \
                      '    if event_type not in (hou.hipFileEventType.AfterSave, hou.hipFileEventType.AfterLoad) and event_type is not None:\n' \
                      '        return\n' \
                      f'    hou.hipFile.setName({repr(mask_path)})\n' \
                      'hou.hipFile.addEventCallback(__fix_hip_env__)\n'

        script += \
            f'print("opening file" + {repr(hippath)})\n' \
            f'hou.hipFile.load({repr(hippath)}, ignore_load_warnings=True)\n' \
            f'node = hou.node({repr(driverpath)})\n'

        for attrname in attr_to_context:
            script += f'hou.setContextOption({repr(attrname)}, {repr(attrs[attrname])})\n'

        if node_specialization_code:
            script += 'node = _that_one_node_specialization_function(node)\n'

        # set fake take to allow editing even locked nodes
        if take_parm_name:
            script += f'base_take_name = node.parm({repr(take_parm_name)}).evalAsString()\n' \
                      f'if base_take_name == "_current_":\n' \
                      f'    base_take = hou.takes.currentTake()\n' \
                      f'else:\n' \
                      f'    base_take = hou.takes.findTake(base_take_name)\n' \
                      f'fake_take = base_take.addChildTake()\n' \
                      f'hou.takes.setCurrentTake(fake_take)\n' \
                      f'fake_take.addParmTuplesFromNode(node)\n'

        for parm_name, value in parms_to_set_before_render.items():
            script += \
                f'if node.parm({repr(parm_name)}).eval() != {repr(value)}:\n' \
                f'    node.parm({repr(parm_name)}).deleteAllKeyframes()\n' \
                f'    node.parm({repr(parm_name)}).set({repr(value)})\n'
        scene_description_path = context.param_value('scene file output').strip()
        if scene_description_path != '':
            script += \
                f'node.parm({repr(scene_file_parm_name)}).deleteAllKeyframes()\n' \
                f'node.parm({repr(scene_file_parm_name)}).set({repr(scene_description_path)})\n'

        script += \
            f'for frame in {repr(frames)}:\n' \
            f'    hou.setFrame(frame)\n' \
            f'    skipped = False\n'
        if context.param_value('skip if exists'):
            script += \
                f'    skipped = os.path.exists(node.parm({repr(scene_file_parm_name)}).evalAsString())\n'
        script += \
            f'    if skipped:\n' \
            f'        print("output file already exists, skipping frame %d" % frame)\n' \
            f'    else:\n' \
            f'        print("rendering frame %d" % frame)\n' \
            f'        node.render(frame_range=(frame, frame))\n'
        if spawnlines:
            script += \
                f'    if {repr(context.param_value("gen for skipped"))} or not skipped:\n' \
                f'{spawnlines}'
        script += \
            f'print("all done!")\n'

        inv = InvocationJob(['hython', ':/work_to_do.py'], env=env)
        inv.set_extra_file('work_to_do.py', script)
        res = ProcessingResult(job=inv)
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



from lifeblood.basenode import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.enums import NodeParameterType
from lifeblood_alicevision_modules.base_node import AlicevisionBaseNode
from pathlib import Path
from typing import Iterable


def node_class():
    return AlicevisionCameraInit


class AlicevisionCameraInit(AlicevisionBaseNode):
    def __init__(self, name):
        super(AlicevisionCameraInit, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('inputFolder', 'Input Images Directory', NodeParameterType.STRING, '')
            ui.add_parameter('output', 'Output Camera', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/CameraInit/cameraInit.sfm")
            ui.add_separator()
            ui.add_parameter('useInternalWhiteBalance', 'Use Internal White Balance', NodeParameterType.BOOL, True)
            ui.add_parameter('defaultFieldOfView', 'Default Field of View', NodeParameterType.FLOAT, 45)
            ui.add_parameter('sensorDatabase', 'Sensor Database', NodeParameterType.STRING, '../../aliceVision/share/aliceVision/cameraSensors.db')

    @classmethod
    def label(cls) -> str:
        return 'AV Camera Init'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'alicevision', 'camera', 'init'

    @classmethod
    def type_name(cls) -> str:
        return 'stock_alicevision_camera_init'

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        output = Path(context.param_value('output').strip())
        srcpath = context.param_value('inputFolder').strip()
        if srcpath == '' or not Path(srcpath).is_absolute():
            raise ProcessingError('image directory path must be absolute and not empty')
        if not output.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')

        sensor_db_path = context.param_value('sensorDatabase').strip()

        output.parent.mkdir(parents=True, exist_ok=True)  # TODO: move this to worker

        args = ['aliceVision_cameraInit', '--verboseLevel', 'info']
        args += ['--defaultFieldOfView', context.param_value('defaultFieldOfView')]
        args += ['--allowSingleView', 1]
        args += ['--sensorDatabase', sensor_db_path]
        args += ['--useInternalWhiteBalance', context.param_value('useInternalWhiteBalance')]
        args += ['--imageFolder', srcpath]
        args += ['--output', output]

        job = InvocationJob([str(x) for x in args])
        res = ProcessingResult(job)
        res.set_attribute('av_camera_sfm', str(output))
        return res
